from typing import Callable, List, Tuple, Dict, Optional, Set

from antlr4 import InputStream, CommonTokenStream
from antlr4.tree.Tree import TerminalNodeImpl

from qrogue.game.logic import Message
from qrogue.game.logic import actors
from qrogue.game.logic.actors import Controllable, Riddle, Robot, BaseBot
from qrogue.game.logic.actors.puzzles import Challenge, boss
from qrogue.game.logic.base import StateVector
from qrogue.game.logic.collectibles import Collectible, MultiCollectible, pickup, Qubit, instruction, \
    Instruction, InstructionManager, CollectibleFactory, OrderedCollectibleFactory
from qrogue.game.logic.collectibles.instruction import HGate
from qrogue.game.target_difficulty import ExplicitTargetDifficulty
from qrogue.game.target_factory import EnemyFactory, EnemyTargetFactory, BossFactory
from qrogue.game.world import tiles
from qrogue.game.world.dungeon_generator import parser_util
from qrogue.game.world.dungeon_generator.dungeon_parser.QrogueDungeonLexer import QrogueDungeonLexer
from qrogue.game.world.dungeon_generator.dungeon_parser.QrogueDungeonParser import QrogueDungeonParser
from qrogue.game.world.dungeon_generator.dungeon_parser.QrogueDungeonVisitor import QrogueDungeonVisitor
from qrogue.game.world.dungeon_generator.generator import DungeonGenerator
from qrogue.game.world.map import CallbackPack, MapMetaData, LessonMap, rooms
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.util import Config, MapConfig, PathConfig, Logger, CommonQuestions, RandomManager, MyRandom, \
    load_help_text, ParserErrorListener, PuzzleGrammarConfig


class QrogueLevelGenerator(DungeonGenerator, QrogueDungeonVisitor):
    __DEFAULT_NUM_OF_RIDDLE_ATTEMPTS = 7
    __ROBOT_NO_GATES = "none"
    __SPAWN_ROOM_ID = "SR"
    __DEFAULT_SPEAKER = Config.examiner_name()  # todo but later in the game it should default to scientist_name()
    __QUICK_MSG_ID = 0
    __QUICK_MSG_PREFIX = "_qumsg"
    __HELP_TEXT_PREFIX = "helptext"

    class _StaticTemplates:
        @staticmethod
        def is_pickup_none(ref: str) -> bool:
            return ref == 'none'

        @staticmethod
        def is_pickup_score(ref: str) -> bool:
            return ref in ['score']

        @staticmethod
        def is_pickup_key(ref: str) -> bool:
            return ref in ['key', 'keys']

        @staticmethod
        def is_pickup_energy(ref: str) -> bool:
            return ref in ['energy']

        @staticmethod
        def is_collectible_gate(ref: str) -> bool:
            return "gate" in ref

        @staticmethod
        def is_dir_north(dir_str: str) -> bool:
            return dir_str == "North"

        @staticmethod
        def is_dir_east(dir_str: str) -> bool:
            return dir_str == "East"

        @staticmethod
        def is_dir_south(dir_str: str) -> bool:
            return dir_str == "South"

        @staticmethod
        def is_dir_west(dir_str: str) -> bool:
            return dir_str == "West"

        @staticmethod
        def get_boss(ref: QrogueDungeonParser.REFERENCE, reward: Collectible, edits: Optional[int],
                     num_of_qubits: int) -> Optional[actors.Boss]:
            ref = parser_util.normalize_reference(ref.getText())
            if ref.startswith("entanglement"):
                qubit_info = ref[len("entanglement"):]
                if len(qubit_info) >= 2:
                    qubits = int(qubit_info[0]), int(qubit_info[1])
                    # increase num_of_qubits if needed
                    num_of_qubits = max(num_of_qubits, max(qubits[0], qubits[1]) + 1)   # +1 because we start at 0
                else:
                    qubits = None
                return boss.EntanglementBoss(reward, edits, num_of_qubits, entangled_qubits=qubits)
            if ref in ["antientangle", "antientanglement"]:
                return boss.AntiEntangleBoss(reward, edits)
            if ref in ["imaginary", "imaginaryboss"]:
                return boss.ImaginaryBoss(reward, edits)
            return None

    @staticmethod
    def __tile_str_to_code(tile_str: str) -> tiles.TileCode:
        if tile_str.isdigit():
            return tiles.TileCode.Enemy

        elif tile_str == parser_util.COLLECTIBLE_TILE:
            return tiles.TileCode.Collectible
        elif tile_str == parser_util.TRIGGER_TILE:
            return tiles.TileCode.Trigger
        elif tile_str == parser_util.MESSAGE_TILE:
            return tiles.TileCode.Message
        elif tile_str == parser_util.ENERGY_TILE:
            return tiles.TileCode.Energy
        elif tile_str == parser_util.RIDDLER_TILE:
            return tiles.TileCode.Riddler
        elif tile_str == parser_util.FLOOR_TILE:
            return tiles.TileCode.Floor
        elif tile_str == parser_util.OBSTACLE_TILE:
            return tiles.TileCode.Obstacle
        else:
            return tiles.TileCode.Invalid

    @staticmethod
    def __tile_code_to_str(tile_code: tiles.TileCode) -> Optional[str]:
        if tile_code is tiles.TileCode.Enemy:
            # todo print warning?
            return "0"
        elif tile_code is tiles.TileCode.Collectible:
            return parser_util.COLLECTIBLE_TILE
        elif tile_code is tiles.TileCode.Trigger:
            return parser_util.TRIGGER_TILE
        elif tile_code is tiles.TileCode.Teleport:
            return parser_util.TELEPORT_TILE
        elif tile_code is tiles.TileCode.Message:
            return parser_util.MESSAGE_TILE
        elif tile_code is tiles.TileCode.Energy:
            return parser_util.ENERGY_TILE
        elif tile_code is tiles.TileCode.Boss:
            return parser_util.BOSS_TILE
        elif tile_code is tiles.TileCode.Riddler:
            return parser_util.RIDDLER_TILE
        elif tile_code is tiles.TileCode.Challenger:
            return parser_util.CHALLENGER_TILE
        elif tile_code is tiles.TileCode.Floor:
            return parser_util.FLOOR_TILE
        else:
            # todo print warning?
            return None

    def __init__(self, check_achievement_callback: Callable[[str], bool],
                 trigger_event_callback: Callable[[str], None],
                 load_map_callback: Callable[[str, Optional[Coordinate]], None],
                 show_message_callback: Callable[[str, str, Optional[bool], Optional[int]], None],
                 callback_pack: CallbackPack):
        super(QrogueLevelGenerator, self).__init__(0, 0)
        self.__check_achievement = check_achievement_callback
        self.__trigger_event = trigger_event_callback
        self.__load_map = load_map_callback
        self.__show_message = show_message_callback
        self.__cbp = callback_pack
        self.__meta_data = MapMetaData(None, None, True, self.__show_description)

        self.__warnings = 0
        self.__level: Optional[LessonMap] = None
        self.__robot: Optional[Robot] = None
        # todo: update visibility of QrogueLevelGenerator to no one can accidentally call a visit-method with uninitialized __rm
        self.__rm: Optional[MyRandom] = None  # is only used to randomize rewards and puzzles, placements are static

        self.__default_speaker = QrogueLevelGenerator.__DEFAULT_SPEAKER
        self.__messages: Dict[str, Message] = {}

        self.__next_target_id = 0
        self.__next_tile_id = 0

        # "collectible factory" refers to "reward pool" in grammar due to the original purpose, simplicity & readability
        self.__collectible_factories: Dict[str, CollectibleFactory] = {}
        self.__default_collectible_factory: Optional[CollectibleFactory] = None

        # "target difficulty" refers to "stv pool" in grammar due to the original purpose, simplicity and readability
        self.__target_difficulties: Dict[str, ExplicitTargetDifficulty] = {}
        self.__default_target_difficulty: Optional[ExplicitTargetDifficulty] = None

        self.__default_enemy_factory: Optional[EnemyFactory] = None  # needed to create default_tile enemies

        self.__hallways_by_id: Dict[str, tiles.Door] = {}  # hw_id -> Door
        self.__doors: Dict[Coordinate, Dict[Coordinate, tiles.Door]] = {}  # for later hallway generation
        self.__entanglement_locks: Set[str] = set()  # stores hw_id of activated entanglement_locks

        self.__enemy_groups_by_room: Dict[str, Dict[int, List[tiles.Enemy]]] = {}  # room_id -> Dict[1-9] -> enemies
        self.__cur_room_id: Optional[str] = None  # needed for enemy groups
        self.__rooms: Dict[str, rooms.CopyAbleRoom] = {}
        self.__spawn_pos: Optional[Coordinate] = None

        # holds references to already created hallways so that neighbors can use it instead of
        # creating their own redundant hallway
        self.__created_hallways: Dict[Coordinate, Dict[Direction, rooms.Hallway]] = {}

    def __show_description(self):
        if self.__meta_data.description:
            ret = self.__meta_data.description.get(self.__check_achievement)
            if ret:
                title, text = ret
                self.__show_message(title, text, None, self.__meta_data.description.position)

    def _warning(self, text: str, loc_details: Optional[str] = None):
        loc_details = "" if loc_details is None else "~" + loc_details
        level_name = "[meta_data uninitialized]" if self.__meta_data is None else self.__meta_data.name
        parser_util.warning(text, f"{level_name}{loc_details}")
        self.__warnings += 1

    def generate(self, seed: int, file_path: str, in_dungeon_folder: bool = True) -> Tuple[Optional[LessonMap], bool]:
        self.__rm = RandomManager.create_new(seed)
        map_data = PathConfig.read_level(file_path, in_dungeon_folder)

        input_stream = InputStream(map_data)
        lexer = QrogueDungeonLexer(input_stream)
        token_stream = CommonTokenStream(lexer)
        parser = QrogueDungeonParser(token_stream)
        parser.addErrorListener(ParserErrorListener())

        try:
            meta_data, room_matrix = self.visit(parser.start())
        except SyntaxError as se:
            Logger.instance().error(str(se), show=False, from_pycui=False)
            return None, False

        # add empty rooms if rows don't have the same width
        max_len = 0
        for row in room_matrix:
            if len(row) > max_len:
                max_len = len(row)
        for row in room_matrix:
            if len(row) < max_len:
                row += [None] * (max_len - len(row))

        if self.__spawn_pos is None:
            raise SyntaxError("No SpawnRoom provided! Make sure to place 'SR' in the layout and if you defined a "
                              "custom SpawnRoom make sure to tag it as (Spawn).")
        self.__level = LessonMap(meta_data, file_path, seed, room_matrix, self.__robot, self.__spawn_pos,
                                 self.__check_achievement, self.__trigger_event)
        return self.__level, True

    def _add_hallway(self, room1: Coordinate, room2: Coordinate, door: tiles.Door):
        if door:  # for simplicity door could be null so we check it here
            if room1 in self.__doors:
                self.__doors[room1][room2] = door
            else:
                self.__doors[room1] = {room2: door}
            if room2 in self.__doors:
                self.__doors[room2][room1] = door
            else:
                self.__doors[room2] = {room1: door}

    def __get_default_tile(self, tile_str: str, enemy_dic: Dict[int, List[tiles.Enemy]],
                           get_entangled_enemies: Callable[[int], List[tiles.Enemy]],
                           update_entangled_groups: Callable[[tiles.Enemy], None]) -> tiles.Tile:
        if tile_str == parser_util.PLACE_HOLDER_TILE:
            return tiles.Floor()

        tile_code = QrogueLevelGenerator.__tile_str_to_code(tile_str)
        if tile_code is tiles.TileCode.Enemy:
            enemy_id = int(tile_str)
            enemy = tiles.Enemy(enemy_id, self.__default_enemy_factory, self.__cbp.start_fight,
                                get_entangled_enemies, update_entangled_groups,
                                self.__rm.get_seed("creating default Enemy"), self._next_tile_id)
            if enemy_id not in enemy_dic:
                enemy_dic[enemy_id] = []
            enemy_dic[enemy_id].append(enemy)
            return enemy

        elif tile_code is tiles.TileCode.Collectible:
            return tiles.Collectible(self.__default_collectible_factory.produce(self.__rm))

        elif tile_code is tiles.TileCode.Energy:
            return tiles.Collectible(pickup.Energy())

        elif tile_code is tiles.TileCode.Riddler:
            stv = self.__default_target_difficulty.create_statevector(self.__robot, self.__rm)
            reward = self.__default_collectible_factory.produce(self.__rm)
            riddle = Riddle(self._next_target_id(), stv, reward, self.__rm.get_seed("creating a default Riddle"),
                            self.__DEFAULT_NUM_OF_RIDDLE_ATTEMPTS)
            return tiles.Riddler(self.__cbp.open_riddle, riddle)

        elif tile_code is tiles.TileCode.Floor:
            return tiles.Floor()
        elif tile_code is tiles.TileCode.Obstacle:
            return tiles.Obstacle()
        else:
            self._warning(f"Unknown tile without default-value specified: {tile_str}. Using a Floor-Tile instead.")
            return tiles.Floor()

    def __get_draw_strategy(self, ctx: QrogueDungeonParser.Draw_strategyContext):
        if ctx:
            return self.visit(ctx)
        return False  # default value is "random"

    def __teleport_callback(self, map_name: str, spawn_pos: Optional[Coordinate]):
        def cb(confirm: int):
            if confirm == 0:
                self.__load_map(map_name, spawn_pos)

        CommonQuestions.GoingBack.ask(cb)

    def __tunnel_callback(self, room_id: str, pos_in_room: Coordinate):
        self.__level.tunnel(self.__spawn_pos, pos_in_room)  # todo implement correctly

    def _next_target_id(self) -> int:
        val = self.__next_target_id
        self.__next_target_id += 1
        return val

    def _next_tile_id(self) -> int:
        val = self.__next_tile_id
        self.__next_tile_id += 1
        return val

    ##### load from references #####

    def __load_collectible_factory(self, reference: QrogueDungeonParser.REFERENCE) -> CollectibleFactory:
        ref = reference.getText()
        if ref in self.__collectible_factories:
            return self.__collectible_factories[ref]

        ref = parser_util.normalize_reference(ref)
        if QrogueLevelGenerator._StaticTemplates.is_pickup_none(ref):
            return CollectibleFactory.empty()
        elif QrogueLevelGenerator._StaticTemplates.is_pickup_score(ref):
            pool = [pickup.Score()]
        elif QrogueLevelGenerator._StaticTemplates.is_pickup_key(ref):
            pool = [pickup.Key()]
        elif QrogueLevelGenerator._StaticTemplates.is_pickup_energy(ref):
            pool = [pickup.Energy()]
        elif QrogueLevelGenerator._StaticTemplates.is_collectible_gate(ref):
            pool = [self.__load_gate(reference)]
        else:
            self._warning(f"Imports not yet supported: {ref}. Choosing from default_reward_factory!")
            # todo implement imports
            return self.__default_collectible_factory
        return CollectibleFactory(pool)

    def __load_target_difficulty(self, reference: QrogueDungeonParser.REFERENCE, allow_default: bool = True) \
            -> Optional[ExplicitTargetDifficulty]:
        ref = reference.getText()
        if ref in self.__target_difficulties:
            return self.__target_difficulties[ref]
        elif allow_default:
            # todo implement imports
            self._warning(f"Imports not yet supported: {ref}. Choosing default_target_difficulty!")
            return self.__default_target_difficulty
        else:
            return None

    def __load_gate(self, reference: QrogueDungeonParser.REFERENCE) -> Instruction:
        ref = parser_util.normalize_reference(reference.getText())
        # remove "gate" if it was added at the end
        if ref.endswith("gate"): ref = ref[:-len("gate")]

        gate = InstructionManager.instruction_from_name(ref)
        if gate is None:
            self._warning(f"Unknown gate reference: {reference}. Returning I Gate instead.")
            return instruction.IGate()
        return gate

    def __load_message(self, reference: QrogueDungeonParser.REFERENCE) -> Message:
        ref = reference.getText()
        if ref in self.__messages:
            return self.__messages[ref]
        norm_ref = parser_util.normalize_reference(ref)
        if norm_ref in self.__messages:
            return self.__messages[norm_ref]
        elif norm_ref.startswith(QrogueLevelGenerator.__HELP_TEXT_PREFIX):
            help_text_type = norm_ref[len(QrogueLevelGenerator.__HELP_TEXT_PREFIX):]
            help_text = load_help_text(help_text_type)
            if help_text:
                # todo check if we really want to prioritize help texts
                return Message.create_with_title(norm_ref, Config.system_name(), help_text, True, None)
        self._warning(f"Unknown text reference: {ref}. Returning \"Message not found!\"")
        return Message.error("Message not found!")

    def __load_hallway(self, hw_id: str) -> Optional[tiles.Door]:
        if hw_id in self.__hallways_by_id:
            return self.__hallways_by_id[hw_id]
        elif hw_id == parser_util.EMPTY_HALLWAY_CODE:
            return None
        elif hw_id == parser_util.DEFAULT_HALLWAY_STR:
            return tiles.Door(Direction.North)
        else:
            # todo implement hallway imports
            return tiles.Door(Direction.North)

    def __load_room(self, room_id: str, x: int, y: int) -> rooms.CopyAbleRoom:
        hw_dic = parser_util.get_hallways(self.__created_hallways, self.__doors, Coordinate(x, y))
        if room_id in self.__rooms:
            room = self.__rooms[room_id]
            if room.type is rooms.AreaType.SpawnRoom:
                spawn_pos = Coordinate(x, y)
                if self.__spawn_pos is not None:
                    self._warning(f"A second SpawnRoom was defined! Ignoring the first one @{self.__spawn_pos} and "
                                  f"using the new one @{spawn_pos} instead.")
                self.__spawn_pos = spawn_pos
            return room.copy(hw_dic)
        elif room_id == self.__SPAWN_ROOM_ID:
            room = rooms.SpawnRoom(lambda: self.__load_map(MapConfig.back_map_string(), None), None,
                                   hw_dic[Direction.North], hw_dic[Direction.East], hw_dic[Direction.South],
                                   hw_dic[Direction.West], place_teleporter=self.__meta_data.has_teleporter)
            spawn_pos = Coordinate(x, y)
            if self.__spawn_pos is not None:
                self._warning(f"A second SpawnRoom was defined! Ignoring the first one @{self.__spawn_pos} and "
                              f"using the new one @{spawn_pos} instead.")
            self.__spawn_pos = spawn_pos
            return room
        elif room_id[0] == '_':
            # todo handle templates
            room = rooms.CustomRoom(rooms.AreaType.Placeholder, None, hw_dic[Direction.North], hw_dic[Direction.East],
                                    hw_dic[Direction.South], hw_dic[Direction.West])
            self.__rooms[room_id] = room
            return room
        else:
            self._warning(f"room_id \"{room_id}\" not specified and imports not yet supported! "
                          "Placing an empty room instead.")
            room = rooms.Placeholder.empty_room(hw_dic)
            self.__rooms[room_id] = room
            return room

    ##### General area

    def visitInteger(self, ctx: QrogueDungeonParser.IntegerContext) -> int:
        return parser_util.parse_integer(ctx)

    def visitComplex_number(self, ctx: QrogueDungeonParser.Complex_numberContext) -> complex:
        return parser_util.parse_complex(ctx)

    def visitDraw_strategy(self, ctx: QrogueDungeonParser.Draw_strategyContext) -> bool:
        """

        :param ctx:
        :return: True if strategy is 'ordered', False if it is 'random'
        """
        return ctx.ORDERED_DRAW() is not None

    ##### Message area ######

    def visitMessage(self, ctx: QrogueDungeonParser.MessageContext) -> Message:
        return parser_util.parse_message(ctx, self.__default_speaker)

    def visitMessages(self, ctx: QrogueDungeonParser.MessagesContext):
        self.__messages.clear()
        if ctx.MSG_SPEAKER():
            self.__default_speaker = parser_util.parse_speaker(ctx, text_index=None)
        else:
            self.__default_speaker = QrogueLevelGenerator.__DEFAULT_SPEAKER
        for msg in ctx.message():
            message = self.visit(msg)
            self.__messages[message.id] = message
        for message in self.__messages.values():
            if message.alt_message_ref and message.alt_message_ref in self.__messages:
                alt_message = self.__messages[message.alt_message_ref]
                try:
                    message.resolve_message_ref(alt_message)
                except ValueError as ve:
                    self._warning("Message-cycle found: " + str(ve), message.id)

    ##### Reward Pool area #####

    def visitCollectible(self, ctx: QrogueDungeonParser.CollectibleContext) -> Optional[Collectible]:
        if ctx.SCORE_LITERAL():
            val = self.visit(ctx.integer())
            return pickup.Score(val)
        elif ctx.KEY_LITERAL():
            val = self.visit(ctx.integer())
            return pickup.Key(val)
        elif ctx.ENERGY_LITERAL():
            val = self.visit(ctx.integer())
            return pickup.Energy(val)
        elif ctx.GATE_LITERAL():
            return self.__load_gate(ctx.REFERENCE())
        elif ctx.QUBIT_LITERAL():
            if ctx.integer():
                return Qubit(self.visit(ctx.integer()))
            else:
                return Qubit(1)
        elif ctx.NONE_LITERAL():
            return None
        else:
            self._warning("No legal collectible specified!")
        return None

    def visitCollectibles(self, ctx: QrogueDungeonParser.CollectiblesContext) -> List[Collectible]:
        collectible_list = []
        for collectible in ctx.collectible():
            collectible_list.append(self.visit(collectible))
        return collectible_list

    def visitReward_pool(self, ctx: QrogueDungeonParser.Reward_poolContext) -> Tuple[str, CollectibleFactory]:
        factory_id = ctx.REFERENCE().getText()
        ordered = self.__get_draw_strategy(ctx.draw_strategy())
        collectible_list = self.visit(ctx.collectibles())
        if ordered:
            return factory_id, OrderedCollectibleFactory(collectible_list)
        else:
            return factory_id, CollectibleFactory(collectible_list)

    def visitDefault_reward_pool(self, ctx: QrogueDungeonParser.Default_reward_poolContext) -> CollectibleFactory:
        if ctx.REFERENCE():  # implicit definition
            return self.__load_collectible_factory(ctx.REFERENCE())

        else:  # explicit definition
            ordered = self.__get_draw_strategy(ctx.draw_strategy())
            collectible_list = self.visit(ctx.collectibles())
            if ordered:
                return OrderedCollectibleFactory(collectible_list)
            else:
                return CollectibleFactory(collectible_list)

    def visitReward_pools(self, ctx: QrogueDungeonParser.Reward_poolsContext) -> None:
        for reward_pool in ctx.reward_pool():
            factory_id, collectible_factory = self.visit(reward_pool)
            self.__collectible_factories[factory_id] = collectible_factory
        self.__default_collectible_factory = self.visit(ctx.default_reward_pool())

    ##### StateVector Pool area #####

    def visitCircuit_gate_single(self, ctx: QrogueDungeonParser.Circuit_gate_singleContext) -> List[Instruction]:
        qubit_index = int(ctx.QUBIT_SPECIFIER().getText()[1:])
        gates: List[Instruction] = []
        for gate_spec in ctx.GATE_SPECIFIER():
            name = gate_spec.getText()[1:]
            gate = InstructionManager.instruction_from_name(name)  # from_name already returns a copy
            if gate is None:
                parser_util.error(f"Invalid gate name: {name}")
                continue
            gate.use_qubit(qubit_index)
            gates.append(gate)
        return gates

    def visitCircuit_gate_multi(self, ctx: QrogueDungeonParser.Circuit_gate_multiContext) -> Optional[Instruction]:
        name = ctx.GATE_SPECIFIER().getText()[1:]
        gate = InstructionManager.instruction_from_name(name)  # from_name already returns a copy
        if gate is None:
            parser_util.error(f"Invalid gate name: {name}")
            return None

        num_provided_qubits = len(ctx.QUBIT_SPECIFIER())
        if gate.num_of_qubits > num_provided_qubits:
            parser_util.error(f"Not enough qubits provided! {gate.num_of_qubits} are needed but only "
                              f"{num_provided_qubits} are specified.")
            return None
        elif gate.num_of_qubits < num_provided_qubits:
            self._warning(f"Too much qubits provided. {gate.num_of_qubits} are needed but {num_provided_qubits} are "
                          f"specified. The last {num_provided_qubits - gate.num_of_qubits} qubits are ignored.", name)

        for qubit_spec in ctx.QUBIT_SPECIFIER():
            qubit = int(qubit_spec.getText()[1:])
            gate.use_qubit(qubit)

        return gate

    def visitCircuit_stv(self, ctx: QrogueDungeonParser.Circuit_stvContext) -> Tuple[List[Instruction], int]:
        num_qubits = parser_util.parse_integer(ctx)

        gates: List[Instruction] = []
        for single_spec in ctx.circuit_gate_single():
            gates += self.visit(single_spec)
        for multi_spec in ctx.circuit_gate_multi():
            gate = self.visit(multi_spec)
            if gate is not None: gates.append(gate)

        return gates, num_qubits

    def visitStv(self, ctx: QrogueDungeonParser.StvContext) -> StateVector:
        if ctx.circuit_stv():
            gates, num_qubits = self.visit(ctx.circuit_stv())
            return Instruction.compute_stv(gates, num_qubits)
        else:
            amplitudes = []
            for cn in ctx.complex_number():
                amplitudes.append(self.visit(cn))
            if not StateVector.check_amplitudes(amplitudes):
                self._warning(f"Invalid amplitudes for StateVector: {amplitudes}! Using 0-only basis state instead.")
                amplitudes = [1] + [0] * (2 ** self.__robot.num_of_qubits - 1)
            return StateVector(amplitudes)

    def visitStv_ref(self, ctx: QrogueDungeonParser.Stv_refContext) -> List[StateVector]:
        if ctx.stv():
            return [self.visit(ctx.stv())]
        else:
            difficulty = self.__load_target_difficulty(ctx.REFERENCE(), allow_default=False)
            if difficulty:
                return difficulty.copy_pool()
            else:
                diff_id = ctx.REFERENCE().getText()
                self._warning(f"Illegal diff_id: {diff_id}. Make sure to only reference previously defined pool ids!")
                return []

    def visitStvs(self, ctx: QrogueDungeonParser.StvsContext) -> List[StateVector]:
        stvs = []
        for stv_ref in ctx.stv_ref():
            stvs += self.visit(stv_ref)
        return stvs

    def visitStv_pool(self, ctx: QrogueDungeonParser.Stv_poolContext) -> Tuple[str, ExplicitTargetDifficulty]:
        diff_id = ctx.REFERENCE(0).getText()
        ordered = self.__get_draw_strategy(ctx.draw_strategy())
        stvs = self.visit(ctx.stvs())

        if ctx.REFERENCE(1):
            reward_factory = self.__load_collectible_factory(ctx.REFERENCE(1))
        else:
            reward_factory = self.__default_collectible_factory

        return diff_id, ExplicitTargetDifficulty(stvs, reward_factory, ordered)

    def visitDefault_stv_pool(self, ctx: QrogueDungeonParser.Default_stv_poolContext) -> ExplicitTargetDifficulty:
        if ctx.REFERENCE():  # implicit definition
            return self.__load_target_difficulty(ctx.REFERENCE())

        else:  # explicit definition
            ordered = self.__get_draw_strategy(ctx.draw_strategy())
            stv_list = self.visit(ctx.stvs())
            return ExplicitTargetDifficulty(stv_list, self.__default_collectible_factory, ordered)

    def visitStv_pools(self, ctx: QrogueDungeonParser.Stv_poolsContext) -> None:
        for stv_pool in ctx.stv_pool():
            diff_id, target_difficulty = self.visit(stv_pool)
            self.__target_difficulties[diff_id] = target_difficulty
        self.__default_target_difficulty = self.visit(ctx.default_stv_pool())
        self.__default_enemy_factory = EnemyTargetFactory(self.__default_target_difficulty,
                                                          next_id_callback=self._next_target_id)

    ##### Hallway area #####

    def visitH_attributes(self, ctx: QrogueDungeonParser.H_attributesContext) -> Tuple[tiles.Door, List[str]]:
        direction = Direction.North  # todo adapt
        one_way_state = tiles.DoorOneWayState.NoOneWay
        if ctx.DIRECTION():
            dir_str = ctx.DIRECTION().getText()
            if QrogueLevelGenerator._StaticTemplates.is_dir_north(dir_str):
                direction = Direction.North
            elif QrogueLevelGenerator._StaticTemplates.is_dir_east(dir_str):
                direction = Direction.East
            elif QrogueLevelGenerator._StaticTemplates.is_dir_south(dir_str):
                direction = Direction.South
            elif QrogueLevelGenerator._StaticTemplates.is_dir_west(dir_str):
                direction = Direction.West
            if ctx.PERMANENT_LITERAL():
                one_way_state = tiles.DoorOneWayState.Permanent
            else:
                one_way_state = tiles.DoorOneWayState.Temporary

        ref_index = 0
        event_id = None
        if ctx.OPEN_LITERAL():
            open_state = tiles.DoorOpenState.Open
        elif ctx.CLOSED_LITERAL():
            open_state = tiles.DoorOpenState.Closed
        elif ctx.LOCKED_LITERAL():
            open_state = tiles.DoorOpenState.KeyLocked
        elif ctx.EVENT_LITERAL():
            if ctx.REFERENCE(ref_index):
                event_id = parser_util.normalize_reference(ctx.REFERENCE(ref_index).getText())
                open_state = tiles.DoorOpenState.EventLocked
                ref_index += 1
            else:
                open_state = tiles.DoorOpenState.Closed
                self._warning("Event lock specified without an event id! Ignoring the lock and placing a closed door "
                              "instead.")
        else:
            open_state = tiles.DoorOpenState.Closed
            self._warning("Invalid hallway attribute: it is neither locked nor opened nor closed!")
        entangled_ids = []
        for hw_id in ctx.HALLWAY_ID():
            entangled_ids.append(hw_id.symbol.text)

        def door_check():
            return self.__check_achievement(event_id)

        door = tiles.Door(direction, open_state, one_way_state, door_check)

        # todo move tutorial and trigger from attributes to hallway?
        if ctx.TUTORIAL_LITERAL():
            message = self.__load_message(ctx.REFERENCE(ref_index))
            door.set_explanation(message)
            ref_index += 1
        if ctx.TRIGGER_LITERAL():
            event_to_trigger = parser_util.normalize_reference(ctx.REFERENCE(ref_index).getText())
            door.set_event(event_to_trigger)
        return door, entangled_ids  # ctx.ENTANGLED_LITERAL() is not None

    def visitHallway(self, ctx: QrogueDungeonParser.HallwayContext) -> Tuple[str, rooms.Hallway]:
        hw_id = ctx.HALLWAY_ID().getText()
        door, entangled_ids = self.visit(ctx.h_attributes())

        if len(entangled_ids) > 0:
            def check_entanglement_lock() -> bool:
                """

                :return: True if this door is locked via entanglement, False otherwise
                """
                for eid in entangled_ids:
                    if eid in self.__entanglement_locks:
                        return True
                self.__entanglement_locks.add(hw_id)
                return False

            door.set_entanglement(check_entanglement_lock)

        return hw_id, door

    def visitHallways(self, ctx: QrogueDungeonParser.HallwaysContext) -> None:
        for hallway_ctx in ctx.hallway():
            hw_id, door = self.visit(hallway_ctx)
            self.__hallways_by_id[hw_id] = door

    ##### Room area #####

    def visitPuzzle_parameter(self, ctx: QrogueDungeonParser.Puzzle_parameterContext) \
            -> Tuple[StateVector, Collectible, Optional[StateVector]]:
        ref_index = 0
        if ctx.stv():
            stv = self.visit(ctx.stv())
        else:
            difficulty = self.__load_target_difficulty(ctx.REFERENCE(ref_index))
            ref_index += 1
            stv = difficulty.create_statevector(self.__robot, self.__rm)

        if ctx.collectible():
            reward = self.visit(ctx.collectible())
        else:
            reward_factory = self.__load_collectible_factory(ctx.REFERENCE(ref_index))
            reward = reward_factory.produce(self.__rm)

        input_stv = self.visit(ctx.input_stv()) if ctx.input_stv() else None

        return stv, reward, input_stv

    def visitInput_stv(self, ctx: QrogueDungeonParser.Input_stvContext) -> StateVector:
        if ctx.REFERENCE():
            difficulty = self.__load_target_difficulty(ctx.REFERENCE())
            return difficulty.create_statevector(self.__robot, self.__rm)
        else:
            return self.visit(ctx.stv())

    def visitRiddle_descriptor(self, ctx: QrogueDungeonParser.Riddle_descriptorContext) -> tiles.Riddler:
        attempts = self.visit(ctx.integer())
        stv, reward, input_stv = self.visit(ctx.puzzle_parameter())
        riddle = Riddle(self._next_target_id(), stv, reward, self.__rm.get_seed("creating a defined Riddle"), attempts,
                        input_stv)
        return tiles.Riddler(self.__cbp.open_riddle, riddle)

    def visitChallenge_descriptor(self, ctx: QrogueDungeonParser.Challenge_descriptorContext) -> tiles.Challenger:
        min_gates = self.visit(ctx.integer(0))
        if ctx.integer(1):
            max_gates = self.visit(ctx.integer(1))
        else:
            max_gates = min_gates

        stv, special_gate, input_stv = self.visitPuzzle_parameter(ctx.puzzle_parameter())
        if input_stv is None:
            # Challenges usually use 0-state targets and non-0 inputs
            input_stv = stv
            stv = StateVector.create_zero_state_vector(stv.num_of_qubits)
            Logger.instance().warn("LevelGenerator found Challenge with 0-state input and non-0 target. Swapping these "
                                   "two for a more consistent Challenge experience.", from_pycui=False)
        if not isinstance(special_gate, Instruction):
            Logger.instance().error("LevelGenerator found Challenge with no gate reward! Using HGate instead.",
                                    from_pycui=False)
            special_gate = HGate()

        # don't specify reward to use the default one
        challenge = Challenge(self._next_target_id(), stv, min_gates, max_gates, input_=input_stv)
        return tiles.Challenger(challenge, special_gate, self.__cbp.open_challenge,
                                lambda title, msg: self.__show_message(title, msg, False, None))

    def visitEnergy_descriptor(self, ctx: QrogueDungeonParser.Energy_descriptorContext) -> tiles.Tile:
        amount = self.visit(ctx.integer())
        return tiles.Energy(amount)

    def visitTrigger_descriptor(self, ctx: QrogueDungeonParser.Trigger_descriptorContext) -> tiles.Trigger:
        if ctx.GLOBAL_ACHIEVEMENT():
            ref = MapConfig.global_event_prefix() + parser_util.normalize_reference(ctx.REFERENCE().getText())
        elif ctx.UNLOCK():
            ref = MapConfig.unlock_prefix() + parser_util.normalize_reference(ctx.REFERENCE().getText())
        else:  # ctx.LEVEL_ACHIEVEMENT() is the default value so we don't have to check for it
            ref = parser_util.normalize_reference(ctx.REFERENCE().getText())

        def callback(_: Direction, __: Controllable):
            self.__trigger_event(ref)

        return tiles.Trigger(callback)

    def visitTeleport_descriptor(self, ctx: QrogueDungeonParser.Teleport_descriptorContext) -> tiles.Teleport:
        if ctx.GLOBAL_TELEPORT():
            ref = parser_util.normalize_reference(ctx.REFERENCE().getText())
            return tiles.Teleport(self.__load_map, ref, None)
        elif ctx.LOCAL_TUNNEL():
            room_id = ctx.ROOM_ID().getText()
            if ctx.integer():
                num = self.visit(ctx.integer())
                x = num % MapConfig.room_width()
                y = int(num / MapConfig.room_height())
            else:
                x, y = MapConfig.room_mid()
            return tiles.Tunnel(self.__tunnel_callback, room_id, Coordinate(x, y))
        else:
            parser_util.error("Invalid Teleport description!")

    def visitT_descriptor(self, ctx: QrogueDungeonParser.T_descriptorContext) -> tiles.Tile:
        if ctx.trigger_descriptor():
            return self.visit(ctx.trigger_descriptor())
        elif ctx.teleport_descriptor():
            return self.visit(ctx.teleport_descriptor())
        else:
            parser_util.error("Invalid T-descriptor! Neither trigger nor teleporter was provided")

    def visitMessage_descriptor(self, ctx: QrogueDungeonParser.Message_descriptorContext) -> tiles.Message:
        if ctx.integer():
            times = self.visit(ctx.integer())
        else:
            times = -1  # = always show
        if ctx.REFERENCE() is None:
            message = Message.create_with_title(f"{self.__QUICK_MSG_PREFIX}{self.__QUICK_MSG_ID}",
                                                title=self.__default_speaker, text=parser_util.text_to_str(ctx),
                                                priority=False)
            self.__QUICK_MSG_ID += 1
        else:
            message = self.__load_message(ctx.REFERENCE())
        return tiles.Message(message, times)

    def visitCollectible_descriptor(self, ctx: QrogueDungeonParser.Collectible_descriptorContext) -> tiles.Collectible:
        if ctx.REFERENCE():
            collectible_factory = self.__load_collectible_factory(ctx.REFERENCE())
            times = 1
            if ctx.integer():
                times = self.visit(ctx.integer())
            if times > 1:
                collectible = MultiCollectible(collectible_factory.produce_multiple(self.__rm, times))
            else:
                collectible = collectible_factory.produce(self.__rm)
        else:
            collectible = self.visit(ctx.collectible())
            if collectible is None:
                self._warning("Wrongly described collectible! Creating one of the default pool instead.")
                collectible = self.__default_collectible_factory.produce(self.__rm)
        return tiles.Collectible(collectible)

    def visitBoss_descriptor(self, ctx: QrogueDungeonParser.Boss_descriptorContext) -> tiles.Boss:
        ref_index = 2 if ctx.GATE_LITERAL() else 1

        # [optional] retrieve the gate that has to be part of the puzzle
        gate = self.__load_gate(ctx.REFERENCE(1)) if ctx.GATE_LITERAL() else None

        # retrieve the specified reward
        if ctx.collectible():
            reward = self.visit(ctx.collectible())
        elif ctx.REFERENCE(ref_index):
            reward = self.__load_collectible_factory(ctx.REFERENCE(ref_index)).produce(self.__rm)
        else:
            # if neither a collectible nor a reference to a reward_factory is given we use the default one
            reward = self.__default_collectible_factory.produce(self.__rm)

        edits = self.visitInteger(ctx.integer())
        boss_actor = QrogueLevelGenerator._StaticTemplates.get_boss(ctx.REFERENCE(0), reward, edits,
                                                                    self.__robot.num_of_qubits)
        if boss_actor is None:
            boss_ref = parser_util.normalize_reference(ctx.REFERENCE(0).getText())
            # check if boss is specified by a dynamic code
            if boss_ref[:len(PuzzleGrammarConfig.boss_code())] == PuzzleGrammarConfig.boss_code():
                code = boss_ref[len(PuzzleGrammarConfig.boss_code()):]
                factory = BossFactory.from_difficulty_code(code, self.__robot, [reward])
            else:
                # produce default boss
                Logger.instance().warn(f"Could not find template for Boss \"{boss_ref}\". Using default "
                                       f"instead.", from_pycui=False)
                factory = BossFactory.default(self.__robot)
            boss_actor = factory.produce(self.__rm, include_gates=[gate] if gate else None)
        return tiles.Boss(boss_actor, self.__cbp.start_boss_fight)

    def visitBoss_puzzle(self, ctx: QrogueDungeonParser.Boss_puzzleContext) -> Tuple[StateVector, StateVector]:
        if ctx.REFERENCE():
            difficulty = self.__load_target_difficulty(ctx.REFERENCE())
            target_stv = difficulty.create_statevector(self.__robot, self.__rm)
        else:
            target_stv = self.visit(ctx.stv())
        input_stv = self.visit(ctx.input_stv())
        return target_stv, input_stv

    def visitEnemy_descriptor(self, ctx: QrogueDungeonParser.Enemy_descriptorContext) -> tiles.Enemy:
        enemy = None
        if self.__cur_room_id:
            room_id = self.__cur_room_id
        else:
            raise NotImplementedError()  # todo better check

        def get_entangled_tiles(id_: int) -> List[tiles.Enemy]:
            if room_id in self.__enemy_groups_by_room:
                room_dic = self.__enemy_groups_by_room[room_id]
                return room_dic[id_]
            else:
                return [enemy]

        def update_entangled_room_groups(new_enemy: tiles.Enemy):
            if room_id not in self.__enemy_groups_by_room:
                self.__enemy_groups_by_room[room_id] = {}
            if enemy_id not in self.__enemy_groups_by_room[room_id]:
                self.__enemy_groups_by_room[room_id][enemy_id] = []
            self.__enemy_groups_by_room[room_id][enemy_id].append(new_enemy)

        enemy_id = int(ctx.DIGIT().getText())

        # reward factory is needed to create the enemy factory so we have to create it first and find out where to look
        if ctx.stv():
            # since a state vector was defined instead of referenced, a reward factory reference can only be the first
            ref_index = 0  # reference occurrence
        else:
            # otherwise a stv was referenced and therefore, a reward factory can only be the second reference
            ref_index = 1

        if ctx.collectible():
            reward = self.visit(ctx.collectible())
            reward_factory = CollectibleFactory([reward])
        elif ctx.REFERENCE(ref_index):
            reward_factory = self.__load_collectible_factory(ctx.REFERENCE(ref_index))
        else:
            # if neither a collectible nor a reference to a reward_factory is given we use the default one
            reward_factory = None

        if ctx.stv():
            stv = self.visit(ctx.stv())
            if reward_factory is None:
                reward_factory = self.__default_collectible_factory  # set reward factory here to not set a custom one
            difficulty = ExplicitTargetDifficulty([stv], reward_factory)
        else:
            # don't use ref_index here because it will always be 0 if present
            difficulty = self.__load_target_difficulty(ctx.REFERENCE(0))

        if ctx.input_stv():
            input_stv = self.visit(ctx.input_stv())
            input_difficulty = ExplicitTargetDifficulty([input_stv])
        else:
            input_difficulty = None

        enemy_factory = EnemyTargetFactory(difficulty, input_difficulty, next_id_callback=self._next_target_id)
        if reward_factory:  # if a reward pool was specified we use it
            enemy_factory.set_custom_reward_factory(reward_factory)
        # else we use the default one either of the loaded difficulty or it already is default_collectible_factory

        enemy = tiles.Enemy(enemy_id, enemy_factory, self.__cbp.start_fight, get_entangled_tiles,
                            update_entangled_room_groups, self.__rm.get_seed("creating defined Enemy"),
                            self._next_tile_id)
        #   update_entangled_room_groups(enemy) # by commenting this the original (copied from) tile is not in the list
        return enemy

    def visitTile_descriptor(self, ctx: QrogueDungeonParser.Tile_descriptorContext) -> tiles.Tile:
        if ctx.t_descriptor():
            tile = self.visit(ctx.t_descriptor())
        elif ctx.enemy_descriptor():
            tile = self.visit(ctx.enemy_descriptor())
        elif ctx.boss_descriptor():
            tile = self.visit(ctx.boss_descriptor())
        elif ctx.collectible_descriptor():
            tile = self.visit(ctx.collectible_descriptor())
        elif ctx.energy_descriptor():
            tile = self.visit(ctx.energy_descriptor())
        elif ctx.riddle_descriptor():
            tile = self.visit(ctx.riddle_descriptor())
        elif ctx.challenge_descriptor():
            tile = self.visit(ctx.challenge_descriptor())
        elif ctx.message_descriptor():
            tile = self.visit(ctx.message_descriptor())
        else:
            self._warning("Invalid tile_descriptor! It is neither enemy, collectible, trigger or energy. "
                          "Returning tiles.Invalid() as consequence.")  # todo: fix message (doesn't list all valid possibilities)
            return tiles.Invalid()

        if isinstance(tile, tiles.WalkTriggerTile):
            ref_index = 0
            if ctx.TUTORIAL_LITERAL():
                msg = self.__load_message(ctx.REFERENCE(ref_index))
                tile.set_explanation(msg)
                ref_index += 1
            if ctx.TRIGGER_LITERAL():
                if ctx.GLOBAL_EVENT_REFERENCE():
                    event_id = parser_util.normalize_reference(ctx.GLOBAL_EVENT_REFERENCE().getText())
                elif ctx.UNLOCK_REFERENCE():
                    event_id = parser_util.normalize_reference(ctx.UNLOCK_REFERENCE().getText())
                else:
                    event_id = parser_util.normalize_reference(ctx.REFERENCE(ref_index).getText())
                tile.set_event(event_id)
        return tile

    def visitTile(self, ctx: QrogueDungeonParser.TileContext) -> str:
        return ctx.getText()

    def visitR_row(self, ctx: QrogueDungeonParser.R_rowContext) -> List[str]:
        row = []
        for tile in ctx.tile():
            row.append(self.visit(tile))
        return row

    def visitR_type(self, ctx: QrogueDungeonParser.R_typeContext) -> rooms.AreaType:
        if ctx.SPAWN_LITERAL():
            return rooms.AreaType.SpawnRoom
        elif ctx.BOSS_LITERAL():
            return rooms.AreaType.BossRoom
        elif ctx.WILD_LITERAL():
            return rooms.AreaType.WildRoom
        elif ctx.RIDDLE_LITERAL():
            return rooms.AreaType.RiddleRoom
        elif ctx.GATE_ROOM_LITERAL():
            return rooms.AreaType.GateRoom
        elif ctx.TREASURE_LITERAL():
            return rooms.AreaType.TreasureRoom
        elif ctx.CHALLENGE_LITERAL():
            return rooms.AreaType.ChallengeRoom
        elif ctx.PAUSE_LITERAL():
            return rooms.AreaType.PauseRoom
        elif ctx.STORY_LITERAL():
            return rooms.AreaType.StoryRoom
        else:
            self._warning(f"Invalid r_type: {ctx.getText()}")
            return rooms.AreaType.Invalid

    def visitR_visibility(self, ctx: QrogueDungeonParser.R_visibilityContext) -> Tuple[bool, bool]:
        visible = False
        foggy = False
        if ctx.VISIBLE_LITERAL():
            visible = True
        elif ctx.FOGGY_LITERAL():
            foggy = True
        return visible, foggy

    def visitR_attributes(self, ctx: QrogueDungeonParser.R_attributesContext) \
            -> Tuple[Tuple[bool, bool], rooms.AreaType]:
        visibility = self.visit(ctx.r_visibility())
        rtype = self.visit(ctx.r_type())
        return visibility, rtype

    def visitRoom_content(self, ctx) -> Tuple[List[str], Dict[str, List[tiles.Tile]]]:
        # place the tiles correctly in the room
        rows = []
        for row in ctx.r_row():
            rows.append(self.visit(row))
        tile_dic = {}
        for descriptor in ctx.tile_descriptor():
            tile = self.visit(descriptor)
            if tile.code is tiles.TileCode.Enemy:
                tile_str = str(tile.eid)
            else:
                tile_str = QrogueLevelGenerator.__tile_code_to_str(tile.code)
            if tile_str in tile_dic:
                tile_dic[tile_str].append(tile)
            else:
                tile_dic[tile_str] = [tile]
        return rows, tile_dic

    def visitRoom(self, ctx: QrogueDungeonParser.RoomContext) -> Tuple[str, rooms.CustomRoom]:
        self.__cur_room_id = ctx.ROOM_ID().getText()
        visibility, room_type = self.visit(ctx.r_attributes())

        rows, tile_dic = self.visit(ctx.room_content())

        # this local method is needed here for not explicitly defined enemies
        # but since we are already in a room we don't have to reference to global data
        enemy_dic = {}

        def get_entangled_enemies(eid: int) -> List[tiles.Enemy]:
            if eid in enemy_dic:
                return enemy_dic[eid]
            else:
                return []

        def update_entangled_room_groups(new_enemy: tiles.Enemy):
            if new_enemy.eid not in enemy_dic:
                enemy_dic[new_enemy.eid] = []
            enemy_dic[new_enemy.eid].append(new_enemy)

        descriptor_indices = {}
        tile_matrix = []
        for row in rows:
            matrix_row = []
            for tile_str in row:
                if tile_str in tile_dic:
                    if tile_str not in descriptor_indices:
                        descriptor_indices[tile_str] = 0
                    index = descriptor_indices[tile_str]

                    # we need to copy it because in case we have less descriptors than tiles we reuse the latest
                    # descriptor
                    tile = tile_dic[tile_str][index].copy()
                    if index + 1 < len(tile_dic[tile_str]):
                        descriptor_indices[tile_str] = index + 1
                elif tile_str == tiles.Teleport.Img() and room_type is rooms.AreaType.SpawnRoom:
                    tile = tiles.Teleport(self.__teleport_callback, MapConfig.back_map_string(), None)
                else:
                    tile = self.__get_default_tile(tile_str, enemy_dic, get_entangled_enemies,
                                                   update_entangled_room_groups)
                matrix_row.append(tile)
            # extended to the needed width with floors
            matrix_row += [tiles.Floor()] * (rooms.Room.INNER_WIDTH - len(matrix_row))
            tile_matrix.append(matrix_row)

        while len(tile_matrix) < rooms.Room.INNER_HEIGHT:
            tile_matrix.append([tiles.Floor()] * rooms.Room.INNER_WIDTH)

        # hallways will be added later
        room = rooms.CustomRoom(room_type, tile_matrix)
        visible, foggy = visibility
        if visible:
            room.make_visible()
        elif foggy:
            room.in_sight()
        return self.__cur_room_id, room

    def visitRooms(self, ctx: QrogueDungeonParser.RoomsContext):
        for room_ctx in ctx.room():
            room_id, room = self.visit(room_ctx)
            self.__rooms[room_id] = room

    ##### Layout area #####

    def visitLayout(self, ctx: QrogueDungeonParser.LayoutContext) -> List[List[Optional[rooms.Room]]]:
        # first setup all hallway connections
        for y, hw_row in enumerate(ctx.l_hallway_row()):
            self.__visitL_hallway_row(hw_row, y)

        room_matrix: List[List[Optional[rooms.Room]]] = []
        for y in range(MapConfig.map_height()):
            row_ctx = ctx.l_room_row(y)
            if row_ctx:
                room_matrix.append(self.__visitL_room_row(row_ctx, y))
            else:
                break
        if ctx.l_room_row(MapConfig.map_height()):
            self._warning(f"Too much room rows specified. Only maps of size ({MapConfig.map_width()}, "
                          f"{MapConfig.map_height()}) supported. Ignoring over-specified rows.")

        return room_matrix

    def __hallway_handling(self, ctx_children: List[TerminalNodeImpl], y: int, direction: Direction):
        x = 0
        for child in ctx_children:
            if parser_util.check_for_overspecified_columns(x, child.symbol.type,
                                                           QrogueDungeonParser.VERTICAL_SEPARATOR):
                self._warning(
                    f"Too much room columns specified. Only maps of size ({MapConfig.map_width()}, "
                    f"{MapConfig.map_height()}) supported. Ignoring over-specified columns.")
                break
            if child.symbol.type == QrogueDungeonParser.HALLWAY_ID:
                hw_id = child.symbol.text
                origin = Coordinate(x, y)
                self._add_hallway(origin, origin + direction, self.__load_hallway(hw_id))
                x += 1
            elif child.symbol.type == QrogueDungeonParser.EMPTY_HALLWAY:
                x += 1

    def __visitL_hallway_row(self, ctx: QrogueDungeonParser.L_hallway_rowContext, y: int) -> None:
        self.__hallway_handling(ctx.children, y, Direction.South)  # connect downwards to the next room row

    def __visitL_room_row(self, ctx: QrogueDungeonParser.L_room_rowContext, y: int) -> List[Optional[rooms.Room]]:
        self.__hallway_handling(ctx.children, y, Direction.East)  # connect to the right to the next room

        row: List[Optional[rooms.Room]] = []
        x = 0
        for child in ctx.children:
            if parser_util.check_for_overspecified_columns(x, child.symbol.type,
                                                           QrogueDungeonParser.VERTICAL_SEPARATOR):
                self._warning(f"Too much room columns specified. Only maps of size ({MapConfig.map_width()}, "
                              f"{MapConfig.map_height()}) supported. Ignoring over-specified columns.")
                break

            if child.symbol.type == QrogueDungeonParser.ROOM_ID:
                room_id = child.symbol.text  # todo make it illegal to have the same room_id twice?
                row.append(self.__load_room(room_id, x, y))
                x += 1
            elif child.symbol.type == QrogueDungeonParser.EMPTY_ROOM:
                row.append(None)
                x += 1
        return row

    ##### Robot area #####

    def visitRobot(self, ctx: QrogueDungeonParser.RobotContext) -> None:
        num_of_qubits = int(ctx.DIGIT().getText())
        gates = []
        for ref in ctx.REFERENCE():     # todo: get rid of this and use NewSaveData's gates instead? but then custom levels would be harder to integrate
            if parser_util.normalize_reference(ref.getText()) == self.__ROBOT_NO_GATES:
                continue
            gates.append(self.__load_gate(ref))  # todo what about pickups?

        integer_index = 0
        circuit_space, backpack_space = None, None
        max_energy, start_energy = None, None

        if ctx.CIRCUIT_SPACE():
            circuit_space = parser_util.parse_integer(ctx.integer(integer_index))
            integer_index += 1
        if ctx.BACKPACK_SPACE():
            backpack_space = parser_util.parse_integer(ctx.integer(integer_index))
            integer_index += 1

        if ctx.MAX_ENERGY():
            max_energy = parser_util.parse_integer(ctx.integer(integer_index))
            integer_index += 1
            if ctx.START_ENERGY():
                start_energy = parser_util.parse_integer(ctx.integer(integer_index))

        self.__robot = BaseBot(self.__cbp.game_over, num_of_qubits, gates, circuit_space, backpack_space, max_energy,
                               start_energy)

    ##### Meta area #####

    def visitMeta(self, ctx: QrogueDungeonParser.MetaContext) -> MapMetaData:
        name = parser_util.text_to_str(ctx) if ctx.TEXT() else None

        # parse optional intro and end message
        body_index, ref_index = 0, 0
        intro_msg, end_msg = None, None
        if ctx.INTRO_MSG():
            if ctx.message_body(body_index):
                title, priority, position, msg = parser_util.parse_message_body(ctx.message_body(body_index),
                                                                                self.__default_speaker)
                intro_msg = Message.create_with_title("_map_description", title, msg, priority, position)
                body_index += 1
            elif ctx.REFERENCE(ref_index):
                intro_msg = self.__load_message(ctx.REFERENCE(ref_index))
                ref_index += 1
        if ctx.END_MSG():
            if ctx.message_body(body_index):
                title, priority, position, msg = parser_util.parse_message_body(ctx.message_body(body_index),
                                                                                self.__default_speaker)
                end_msg = Message.create_with_title("_map_ending", title, msg, priority, position)
            elif ctx.REFERENCE(ref_index):
                end_msg = self.__load_message(ctx.REFERENCE(ref_index))

        return MapMetaData(name, intro_msg, ctx.NO_TELEPORTER() is None, self.__show_description,
                           ctx.SHOW_INDIV_QUBITS() is not None, end_msg)

    ##### Start area #####

    def visitStart(self, ctx: QrogueDungeonParser.StartContext) -> Tuple[MapMetaData, List[List[Optional[rooms.Room]]]]:
        # prepare messages (needs to be done first since metadata might reference it
        self.visit(ctx.messages())

        # retrieve the map's meta data
        self.__meta_data = self.visit(ctx.meta())
        Logger.instance().info(f"Building map {self.__meta_data.name}", from_pycui=False)

        # prepare the robot
        self.visit(ctx.robot())

        # prepare reward pools first because they are standalone
        self.visit(ctx.reward_pools())
        # prepare state vector pools next because they can only reference reward pools
        self.visit(ctx.stv_pools())

        # prepare hallways next because they are standalone but thematically more connected to rooms
        self.visit(ctx.hallways())
        # now prepare the rooms because they can reference everything above
        self.visit(ctx.rooms())

        # for the last step we retrieve the room matrix from layout
        return self.__meta_data, self.visit(ctx.layout())



# exporting
from .learnables import WFCLearnMatrix, LearnableMap, LearnableRoom
from .wave_function import WaveFunction
from .wfc_generator import WFCGenerator, WFCLayoutGenerator, WFCRoomGenerator, WFCEmptyRoomGenerator
from .wfc_learner import WFCLearner
from .wfc_manager import WFCManager

# importing
# +world.map
# +world.map.rooms
# +world.navigation
# +world.rooms
# +world.tiles
# +util



from abc import abstractmethod, ABC
from typing import List, Callable, Optional, Dict, Union, Tuple

import qrogue.game.logic.collectibles.instruction as gates
from qrogue.game.logic import PuzzleGenerator
from qrogue.game.logic.actors import Challenge
from qrogue.game.logic.actors.controllables import Robot
from qrogue.game.logic.actors.puzzles import Enemy, Target, Boss
from qrogue.game.logic.base import StateVector
from qrogue.game.logic.collectibles import Collectible, CollectibleFactory, Instruction, QuantumFuser
from qrogue.game.target_difficulty import ExplicitTargetDifficulty
from qrogue.util import Logger, MyRandom, Config, StvDifficulty, DifficultyType


class EnemyFactory(ABC):
    @staticmethod
    def _fallback_enemy(id_: Optional[int], eid: Optional[int], num_of_qubits: Optional[int],
                        reward: Optional[Collectible], input_stv: Optional[StateVector] = None) -> Enemy:
        if id_ is None: id_ = -1
        if eid is None: eid = 0
        if num_of_qubits is None: num_of_qubits = 2
        return Enemy(id_, eid, StateVector.create_zero_state_vector(num_of_qubits), reward, input_stv)

    @abstractmethod
    def robot_based(self) -> bool:
        pass

    @abstractmethod
    def produce_enemy(self, rm: MyRandom, eid: int,
                      data: Union[Robot, Tuple[Optional[List[Instruction]], Optional[List[Instruction]]]]) -> Enemy:
        pass


class EnemyTargetFactory(EnemyFactory):
    """
    This class produces enemies (actors) with a certain difficulty.
    It is used by enemy tiles to trigger a fight.
    """

    def __init__(self, target_difficulty: ExplicitTargetDifficulty,
                 input_difficulty: Optional[ExplicitTargetDifficulty] = None,
                 next_id_callback: Optional[Callable[[], int]] = None):
        """

        :param target_difficulty: difficulty of the enemy target we produce
        :param input_difficulty: difficulty of the input we produce
        """
        super().__init__()
        if next_id_callback is None:
            self.__next_id = 0

            def _next_id() -> int:
                val = self.__next_id
                self.__next_id += 1
                return val

            self._next_id = _next_id
        else:
            self._next_id = next_id_callback
        self.__target_difficulty = target_difficulty
        self.__input_difficulty = input_difficulty
        self.__custom_reward_factory = None

    def robot_based(self) -> bool:
        return True

    def set_custom_reward_factory(self, factory: CollectibleFactory):
        self.__custom_reward_factory = factory

    def produce_enemy(self, rm: MyRandom, eid: int,
                      data: Union[Robot, Tuple[Optional[List[Instruction]], Optional[List[Instruction]]]]) -> Enemy:
        """
        Creates an enemy based on the number of qubits the provided robot has.

        :param rm:
        :param eid: id in [0, 9] to calculate certain properties
        :param data: Robot
        :return: a freshly created enemy
        """
        if self.__custom_reward_factory:
            reward = self.__custom_reward_factory.produce(rm)
        elif self.__target_difficulty.has_reward_factory:
            reward = self.__target_difficulty.produce_reward(rm)
        else:
            reward = None

        if isinstance(data, Robot):
            robot: Robot = data
            target_stv = self.__target_difficulty.create_statevector(robot, rm)
            if self.__input_difficulty is None:
                input_stv = None
            else:
                max_rerolls = 10
                input_stv = self.__input_difficulty.create_statevector(robot, rm)

                while input_stv.get_diff(target_stv).is_zero and max_rerolls > 0:
                    input_stv = self.__input_difficulty.create_statevector(robot, rm)
                    max_rerolls -= 1

                # if max_rerolls > 0 we know that the loop above terminated because the vectors are not the same -> done
                if max_rerolls <= 0 and input_stv.get_diff(target_stv).is_zero:
                    # try re-rolling the target
                    max_rerolls = 10
                    target_stv = self.__target_difficulty.create_statevector(robot, rm)
                    while input_stv.get_diff(target_stv).is_zero and max_rerolls > 0:
                        target_stv = self.__target_difficulty.create_statevector(robot, rm)
                        max_rerolls -= 1

                    if max_rerolls <= 0 and input_stv.get_diff(target_stv).is_zero:
                        inst_text = "; ".join([str(inst) for inst in robot.instructions])
                        Logger.instance().warn(f"Couldn't re-roll input and target to be different! {inst_text}",
                                               from_pycui=False)
            return Enemy(self._next_id(), eid, target_stv, reward, input_=input_stv)

        Config.check_reachability("EnemyTargetFactory.produce_enemy(Tuple)")
        Logger.instance().error(f"Invalid parameter data! Expected Robot but got \"{type(data)}\"")
        return EnemyFactory._fallback_enemy(self._next_id(), eid, None, reward)


class PuzzleFactory(ABC):
    def __init__(self, difficulty: StvDifficulty, num_of_qubits: int, circuit_space: int,
                 get_available_gates_callback: Callable[[], List[Instruction]],
                 next_id_callback: Optional[Callable[[], int]] = None):
        self.__difficulty = difficulty
        self.__num_of_qubits = num_of_qubits
        self.__circuit_space = circuit_space
        self.__get_available_gates = get_available_gates_callback
        self.__next_id = 0

        if next_id_callback is None:
            self.__next_id = 0

            def _next_id() -> int:
                val = self.__next_id
                self.__next_id += 1
                return val

            self._next_id = _next_id
        else:
            self._next_id = next_id_callback

    @property
    def _difficulty(self) -> StvDifficulty:
        return self.__difficulty

    @property
    def _num_of_qubits(self) -> int:
        return self.__num_of_qubits

    @property
    def _circuit_space(self) -> int:
        return self.__circuit_space

    def _get_available_gates(self) -> List[Instruction]:
        return self.__get_available_gates()

    def _prepare_from_gates(self, rm: MyRandom, include_gates: List[Instruction], force_num_of_gates: bool,
                            inverse: bool = False) -> List[Instruction]:
        return PuzzleGenerator.prepare_from_gates(rm, self.__num_of_qubits, self.__circuit_space, self.__difficulty,
                                                  self.__get_available_gates(), include_gates, force_num_of_gates,
                                                  inverse)

    @abstractmethod
    def produce(self, rm: MyRandom, include_gates: Optional[List[Instruction]],
                input_gates: Optional[List[Instruction]] = None) -> Target:
        pass


class NewEnemyFactory(PuzzleFactory, EnemyFactory):
    def __init__(self, difficulty: StvDifficulty, num_of_qubits: int, circuit_space,
                 get_available_gates_callback: Callable[[], List[Instruction]],
                 input_data: Union[bool, List[Instruction]], reward_pool: Optional[List[Optional[Collectible]]] = None,
                 next_id_callback: Optional[Callable[[], int]] = None):
        super().__init__(difficulty, num_of_qubits, circuit_space, get_available_gates_callback, next_id_callback)
        if isinstance(input_data, bool):
            self.__rotate_input = input_data
            self.__input_gates = None
        else:
            self.__rotate_input = False
            self.__input_gates = input_data
        self.__reward_pool = reward_pool

    def robot_based(self) -> bool:
        return False

    def produce_enemy(self, rm: MyRandom, eid: int,
                      data: Union[Robot, Tuple[Optional[List[Instruction]], Optional[List[Instruction]]]]) -> Enemy:
        if isinstance(data, Tuple):
            include_gates, input_gates = data
            return self.produce(rm, include_gates, input_gates, eid)

        Config.check_reachability("NewEnemyFactory.produce_enemy(Robot)")
        Logger.instance().error(f"Invalid parameter data! Expected Tuple[List[Instruction], List[Instruction] but got "
                                f"\"{type(data)}\"")
        robot: Robot = data
        return EnemyFactory._fallback_enemy(None, eid, robot.num_of_qubits, None)

    def produce(self, rm: MyRandom, include_gates: Optional[List[Instruction]],
                input_gates: Optional[List[Instruction]] = None, eid: Optional[int] = None) -> Enemy:
        eid = 0 if eid is None else eid

        # Note (0.1.09.24): include_gates and input_gates seem to be None for all calls

        # prepare gates for stv computation
        if input_gates is not None or self.__input_gates is not None:
            # just use the provided gates
            gates_for_input = \
                PuzzleGenerator.prepare_single_layer_gates(rm, self._num_of_qubits, self._circuit_space,
                                                           self._difficulty, self.__input_gates.copy())
        elif self.__rotate_input:
            # use a random rotation
            gates_for_input = PuzzleGenerator.prepare_rotation_gates(rm, self._num_of_qubits, self._circuit_space,
                                                                     self._difficulty)
        else:
            # use no create a 0-state stv
            gates_for_input = []
        gates_for_target = self._prepare_from_gates(rm, include_gates, force_num_of_gates=False)

        target_stv = Instruction.compute_stv(gates_for_input + gates_for_target, self._num_of_qubits)
        # don't reverse this StV because the used gates are used as-is in target_stv and don't have to be "countered"
        #  (i.e., reversed by the player)
        input_stv = Instruction.compute_stv(gates_for_input, self._num_of_qubits, inverse=False)
        reward = None if self.__reward_pool is None \
            else rm.get_element(self.__reward_pool, msg="EnemyFactory.produce()@reward")

        return Enemy(self._next_id(), eid, target_stv, reward, input_stv)


class ChallengeFactory(PuzzleFactory):
    @staticmethod
    def from_robot(difficulty: StvDifficulty, robot: Robot, next_id_callback: Optional[Callable[[], int]] = None) \
            -> "ChallengeFactory":
        available_gates = robot.get_available_instructions()
        return ChallengeFactory(difficulty, robot.num_of_qubits, robot.circuit_space, available_gates, next_id_callback)

    @staticmethod
    def from_difficulty_code(code: str, robot: Robot) -> "BossFactory":
        difficulty = StvDifficulty.from_difficulty_code(code)
        assert difficulty.level >= 0, f"Code \"{code}\" produced invalid level: {difficulty.level} >= 0 is False!"
        return BossFactory.from_robot(difficulty, robot)

    def __init__(self, difficulty: StvDifficulty, num_of_qubits: int, circuit_space: int,
                 available_gates: List[Instruction], next_id_callback: Optional[Callable[[], int]] = None):
        super().__init__(difficulty, num_of_qubits, circuit_space, lambda: available_gates, next_id_callback)

    def produce(self, rm: MyRandom, include_gates: Optional[List[Instruction]],
                input_gates: Optional[List[Instruction]] = None) -> Challenge:
        assert len(include_gates) == 1, "Did not provide exactly 1 gate to include in Challenge!"

        input_gates = self._prepare_from_gates(rm, include_gates, force_num_of_gates=False, inverse=True)
        input_stv = Instruction.compute_stv(input_gates, self._num_of_qubits, inverse=True)
        target_stv = StateVector.create_zero_state_vector(self._num_of_qubits)
        # don't specify reward to use the default one
        return Challenge(self._next_id(), target_stv, len(input_gates), len(input_gates), input_=input_stv)


class BossFactory(PuzzleFactory):
    __LEVELED_REWARD_POOLS: Dict[int, List[Collectible]] = {
        0: [gates.XGate(), gates.CXGate(), gates.HGate()],
        1: [gates.XGate(), gates.CXGate(), gates.HGate()],
        2: [gates.HGate(), gates.CXGate(), gates.SwapGate()],
        3: [gates.HGate(), gates.SGate()],
        4: [gates.SGate(), gates.YGate(), gates.ZGate()],
        5: [gates.RYGate(), gates.RZGate()],
        6: [gates.HGate()],
    }

    @staticmethod
    def validate() -> bool:
        if len(BossFactory.__LEVELED_REWARD_POOLS) != StvDifficulty.num_of_difficulty_levels():
            return False
        for i in range(StvDifficulty.num_of_difficulty_levels()):
            if i + StvDifficulty.min_difficulty_level() not in BossFactory.__LEVELED_REWARD_POOLS:
                return False
        return True

    @staticmethod
    def get_leveled_rewards(level: int) -> List[Collectible]:
        assert StvDifficulty.min_difficulty_level() <= level <= StvDifficulty.max_difficulty_level(), \
            "Invalid reward level: " \
            f"{StvDifficulty.min_difficulty_level()} <= {level} <= {StvDifficulty.max_difficulty_level()} is False!"
        return BossFactory.__LEVELED_REWARD_POOLS[level].copy()

    @staticmethod
    def from_robot(difficulty: StvDifficulty, robot: Robot, reward_pool: Optional[List[Collectible]] = None,
                   next_id_callback: Optional[Callable[[], int]] = None) -> "BossFactory":
        available_gates = robot.get_available_instructions()
        if reward_pool is None:
            reward_pool = BossFactory.__LEVELED_REWARD_POOLS[difficulty.level]
        return BossFactory(difficulty, robot.num_of_qubits, robot.circuit_space, available_gates, reward_pool,
                           next_id_callback)

    @staticmethod
    def from_difficulty_code(code: str, robot: Robot, reward_pool: Optional[List[Collectible]] = None) -> "BossFactory":
        difficulty = StvDifficulty.from_difficulty_code(code)
        assert difficulty.level >= 0, f"Code \"{code}\" produced invalid level: {difficulty.level} >= 0 is False!"
        return BossFactory.from_robot(difficulty, robot, reward_pool)

    @staticmethod
    def from_difficulty_level(level: int, robot: Robot, reward_pool: Optional[List[Collectible]] = None) \
            -> "BossFactory":
        difficulty = StvDifficulty.from_difficulty_level(level)
        return BossFactory.from_robot(difficulty, robot, reward_pool)

    @staticmethod
    def default(robot: Robot) -> "BossFactory":
        return BossFactory.from_difficulty_level(1, robot)

    def __init__(self, difficulty: StvDifficulty, num_of_qubits: int, circuit_space: int,
                 available_gates: List[Instruction], reward_pool: List[Collectible],
                 next_id_callback: Optional[Callable[[], int]] = None):
        super().__init__(difficulty, num_of_qubits, circuit_space, lambda: available_gates, next_id_callback)
        self.__reward_pool = reward_pool

    def produce(self, rm: MyRandom, include_gates: Optional[List[Instruction]],
                input_gates: Optional[List[Instruction]] = None) -> Boss:
        """
        :param rm: used for randomization during the puzzle generation
        :param include_gates: a list of gates that will be included in the puzzle additionally to the available gates
        :param input_gates: a list of additional gates used to compute the input stv (don't have to be available to
                            the solving robot)
        """
        # prepare gates to compute target_stv
        ############################
        gates_for_target = self._prepare_from_gates(rm, include_gates, force_num_of_gates=True)
        ############################

        # prepare input_stv either based on input_gates or difficulty
        ############################
        if input_gates is None:
            # prepare gates used for input_stv based on difficulty
            gates_for_input = PuzzleGenerator.prepare_rotation_gates(rm, self._num_of_qubits, self._circuit_space,
                                                                     self._difficulty)
        else:
            qubit_count = [0] * self._num_of_qubits
            qubits = list(range(self._num_of_qubits))
            # prepare gates used for input_stv based on provided input_gates
            Config.check_reachability("BossFactory.produce() with input_gates")
            gates_for_input: List[Instruction] = []
            # use the provided input gates (copy because otherwise the used gates could be changed from the outside)
            for gate in input_gates.copy():
                if BossFactory.__prepare_gate(rm, self._circuit_space, gate, qubit_count, qubits):
                    gates_for_input.append(gate)
        input_stv = Instruction.compute_stv(gates_for_input, self._num_of_qubits)
        ############################

        # generate target_stv based on selected gates and input
        ############################
        # to make sure that the target can be reached from the randomly or manually generated input state,
        #  gates_for_input is prepended to gates_for_target such that the difference between target and input is solely
        #  the previously generated gates_for_target
        target_stv = Instruction.compute_stv(gates_for_input + gates_for_target, self._num_of_qubits)
        ############################

        # pick a reward
        ############################
        reward = QuantumFuser()     #rm.get_element(self.__reward_pool, msg="BossFactory_reward")
        ############################

        edits = self._difficulty.get_absolute_value(DifficultyType.BonusEditRatio, self._num_of_qubits,
                                                    self._circuit_space)
        return Boss(self._next_id(), target_stv, input_stv, reward, edits)

    @staticmethod
    def __prepare_gate(rm: MyRandom, circuit_space: int, gate: Instruction, qubit_count: List[int], qubits: List[int]) \
            -> bool:
        # todo: qubit_count is redundant if we can only place one gate per column
        #  (implying qubit_count can never be > circuit_space if number of placed gates is <= circuit_space)
        gate_qubits = qubits.copy()
        while len(gate_qubits) > 0:
            qubit = rm.get_element(gate_qubits, remove=True, msg="BossFactory_selectQubit")
            qubit_count[qubit] += 1
            if qubit_count[qubit] >= circuit_space:
                qubits.remove(qubit)
            if not gate.use_qubit(qubit):
                return True
        return False



# Generated from D:/Workspaces/pycharm-workspace/Qrogue/qrogue/management/save_grammar/SaveData.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,51,157,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        2,14,7,14,2,15,7,15,2,16,7,16,2,17,7,17,1,0,1,0,1,0,3,0,40,8,0,1,
        0,1,0,1,0,1,0,1,0,1,0,1,1,1,1,1,1,1,2,1,2,1,2,1,3,1,3,1,3,1,3,1,
        4,1,4,5,4,60,8,4,10,4,12,4,63,9,4,1,5,1,5,1,5,3,5,68,8,5,1,6,1,6,
        1,6,1,6,1,6,1,6,1,6,5,6,77,8,6,10,6,12,6,80,9,6,1,6,1,6,1,7,1,7,
        1,7,1,7,1,7,5,7,89,8,7,10,7,12,7,92,9,7,1,7,1,7,1,8,1,8,5,8,98,8,
        8,10,8,12,8,101,9,8,1,9,1,9,1,9,1,9,1,9,1,9,1,10,1,10,1,10,1,10,
        1,11,1,11,1,12,1,12,5,12,117,8,12,10,12,12,12,120,9,12,1,13,1,13,
        1,13,1,13,1,14,1,14,5,14,128,8,14,10,14,12,14,131,9,14,1,15,1,15,
        1,15,1,15,1,15,1,15,1,15,1,15,1,16,1,16,1,17,3,17,144,8,17,1,17,
        1,17,1,17,3,17,149,8,17,1,17,1,17,3,17,153,8,17,3,17,155,8,17,1,
        17,0,0,18,0,2,4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,34,0,1,2,
        0,36,37,46,46,150,0,36,1,0,0,0,2,47,1,0,0,0,4,50,1,0,0,0,6,53,1,
        0,0,0,8,57,1,0,0,0,10,67,1,0,0,0,12,69,1,0,0,0,14,83,1,0,0,0,16,
        95,1,0,0,0,18,102,1,0,0,0,20,108,1,0,0,0,22,112,1,0,0,0,24,114,1,
        0,0,0,26,121,1,0,0,0,28,125,1,0,0,0,30,132,1,0,0,0,32,140,1,0,0,
        0,34,143,1,0,0,0,36,37,5,28,0,0,37,39,3,2,1,0,38,40,3,6,3,0,39,38,
        1,0,0,0,39,40,1,0,0,0,40,41,1,0,0,0,41,42,3,8,4,0,42,43,3,16,8,0,
        43,44,3,24,12,0,44,45,3,28,14,0,45,46,5,29,0,0,46,1,1,0,0,0,47,48,
        5,19,0,0,48,49,5,20,0,0,49,3,1,0,0,0,50,51,3,22,11,0,51,52,5,17,
        0,0,52,5,1,0,0,0,53,54,5,11,0,0,54,55,5,16,0,0,55,56,3,22,11,0,56,
        7,1,0,0,0,57,61,5,12,0,0,58,60,3,10,5,0,59,58,1,0,0,0,60,63,1,0,
        0,0,61,59,1,0,0,0,61,62,1,0,0,0,62,9,1,0,0,0,63,61,1,0,0,0,64,68,
        5,22,0,0,65,66,5,18,0,0,66,68,3,12,6,0,67,64,1,0,0,0,67,65,1,0,0,
        0,68,11,1,0,0,0,69,70,5,1,0,0,70,71,5,44,0,0,71,72,3,22,11,0,72,
        73,5,2,0,0,73,78,3,14,7,0,74,75,5,32,0,0,75,77,3,14,7,0,76,74,1,
        0,0,0,77,80,1,0,0,0,78,76,1,0,0,0,78,79,1,0,0,0,79,81,1,0,0,0,80,
        78,1,0,0,0,81,82,5,3,0,0,82,13,1,0,0,0,83,84,5,22,0,0,84,85,5,4,
        0,0,85,90,3,22,11,0,86,87,5,32,0,0,87,89,3,22,11,0,88,86,1,0,0,0,
        89,92,1,0,0,0,90,88,1,0,0,0,90,91,1,0,0,0,91,93,1,0,0,0,92,90,1,
        0,0,0,93,94,5,5,0,0,94,15,1,0,0,0,95,99,5,13,0,0,96,98,3,18,9,0,
        97,96,1,0,0,0,98,101,1,0,0,0,99,97,1,0,0,0,99,100,1,0,0,0,100,17,
        1,0,0,0,101,99,1,0,0,0,102,103,5,24,0,0,103,104,5,6,0,0,104,105,
        3,2,1,0,105,106,3,4,2,0,106,107,3,20,10,0,107,19,1,0,0,0,108,109,
        5,7,0,0,109,110,5,8,0,0,110,111,3,22,11,0,111,21,1,0,0,0,112,113,
        5,21,0,0,113,23,1,0,0,0,114,118,5,14,0,0,115,117,3,26,13,0,116,115,
        1,0,0,0,117,120,1,0,0,0,118,116,1,0,0,0,118,119,1,0,0,0,119,25,1,
        0,0,0,120,118,1,0,0,0,121,122,5,22,0,0,122,123,5,6,0,0,123,124,3,
        2,1,0,124,27,1,0,0,0,125,129,5,15,0,0,126,128,3,30,15,0,127,126,
        1,0,0,0,128,131,1,0,0,0,129,127,1,0,0,0,129,130,1,0,0,0,130,29,1,
        0,0,0,131,129,1,0,0,0,132,133,5,22,0,0,133,134,5,6,0,0,134,135,3,
        2,1,0,135,136,3,20,10,0,136,137,5,9,0,0,137,138,5,10,0,0,138,139,
        3,22,11,0,139,31,1,0,0,0,140,141,7,0,0,0,141,33,1,0,0,0,142,144,
        5,40,0,0,143,142,1,0,0,0,143,144,1,0,0,0,144,154,1,0,0,0,145,155,
        5,39,0,0,146,149,3,32,16,0,147,149,5,38,0,0,148,146,1,0,0,0,148,
        147,1,0,0,0,149,152,1,0,0,0,150,151,5,40,0,0,151,153,5,39,0,0,152,
        150,1,0,0,0,152,153,1,0,0,0,153,155,1,0,0,0,154,145,1,0,0,0,154,
        148,1,0,0,0,155,35,1,0,0,0,12,39,61,67,78,90,99,118,129,143,148,
        152,154
    ]

class SaveDataParser ( Parser ):

    grammarFileName = "SaveData.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'{'", "':'", "'}'", "'('", "')'", "'@'", 
                     "'Score'", "'='", "'out'", "'of'", "'[INVENTORY]'", 
                     "'[GATES]'", "'[LEVELS]'", "'[UNLOCKS]'", "'[ACHIEVEMENTS]'", 
                     "'QuantumFuser'", "'seconds'", "'combined'", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'tutorial'", "'trigger'", 
                     "'Qrogue<'", "'>Qrogue'", "'~'", "'|'", "','", "'#'", 
                     "'..'", "'__'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "INVENTORY_HEADER", 
                      "GATES_HEADER", "LEVELS_HEADER", "UNLOCKS_HEADER", 
                      "ACHIEVEMENTS_HEADER", "QUANTUM_FUSER", "DURATION_UNIT", 
                      "COMBINED_GATE", "DATE", "TIME", "VALUE", "NAME_STD", 
                      "NAME_SYMBOL", "NAME_SPECIAL", "DIRECTION", "TUTORIAL_LITERAL", 
                      "TRIGGER_LITERAL", "HEADER", "ENDER", "HORIZONTAL_SEPARATOR", 
                      "VERTICAL_SEPARATOR", "LIST_SEPARATOR", "WALL", "EMPTY_HALLWAY", 
                      "EMPTY_ROOM", "DIGIT", "INTEGER", "FLOAT", "IMAG_NUMBER", 
                      "SIGN", "CHARACTER_LOW", "CHARACTER_UP", "CHARACTER", 
                      "TEXT", "ROOM_ID", "HALLWAY_ID", "REFERENCE", "WS", 
                      "UNIVERSAL_SEPARATOR", "COMMENT", "LINE_COMMENT" ]

    RULE_start = 0
    RULE_date_time = 1
    RULE_duration = 2
    RULE_inventory = 3
    RULE_gates = 4
    RULE_gate = 5
    RULE_combined_gate = 6
    RULE_background_gate = 7
    RULE_levels = 8
    RULE_level = 9
    RULE_score = 10
    RULE_value = 11
    RULE_unlocks = 12
    RULE_unlock = 13
    RULE_achievements = 14
    RULE_achievement = 15
    RULE_integer = 16
    RULE_complex_number = 17

    ruleNames =  [ "start", "date_time", "duration", "inventory", "gates", 
                   "gate", "combined_gate", "background_gate", "levels", 
                   "level", "score", "value", "unlocks", "unlock", "achievements", 
                   "achievement", "integer", "complex_number" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    INVENTORY_HEADER=11
    GATES_HEADER=12
    LEVELS_HEADER=13
    UNLOCKS_HEADER=14
    ACHIEVEMENTS_HEADER=15
    QUANTUM_FUSER=16
    DURATION_UNIT=17
    COMBINED_GATE=18
    DATE=19
    TIME=20
    VALUE=21
    NAME_STD=22
    NAME_SYMBOL=23
    NAME_SPECIAL=24
    DIRECTION=25
    TUTORIAL_LITERAL=26
    TRIGGER_LITERAL=27
    HEADER=28
    ENDER=29
    HORIZONTAL_SEPARATOR=30
    VERTICAL_SEPARATOR=31
    LIST_SEPARATOR=32
    WALL=33
    EMPTY_HALLWAY=34
    EMPTY_ROOM=35
    DIGIT=36
    INTEGER=37
    FLOAT=38
    IMAG_NUMBER=39
    SIGN=40
    CHARACTER_LOW=41
    CHARACTER_UP=42
    CHARACTER=43
    TEXT=44
    ROOM_ID=45
    HALLWAY_ID=46
    REFERENCE=47
    WS=48
    UNIVERSAL_SEPARATOR=49
    COMMENT=50
    LINE_COMMENT=51

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class StartContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def HEADER(self):
            return self.getToken(SaveDataParser.HEADER, 0)

        def date_time(self):
            return self.getTypedRuleContext(SaveDataParser.Date_timeContext,0)


        def gates(self):
            return self.getTypedRuleContext(SaveDataParser.GatesContext,0)


        def levels(self):
            return self.getTypedRuleContext(SaveDataParser.LevelsContext,0)


        def unlocks(self):
            return self.getTypedRuleContext(SaveDataParser.UnlocksContext,0)


        def achievements(self):
            return self.getTypedRuleContext(SaveDataParser.AchievementsContext,0)


        def ENDER(self):
            return self.getToken(SaveDataParser.ENDER, 0)

        def inventory(self):
            return self.getTypedRuleContext(SaveDataParser.InventoryContext,0)


        def getRuleIndex(self):
            return SaveDataParser.RULE_start

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStart" ):
                listener.enterStart(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStart" ):
                listener.exitStart(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStart" ):
                return visitor.visitStart(self)
            else:
                return visitor.visitChildren(self)




    def start(self):

        localctx = SaveDataParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 36
            self.match(SaveDataParser.HEADER)
            self.state = 37
            self.date_time()
            self.state = 39
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==11:
                self.state = 38
                self.inventory()


            self.state = 41
            self.gates()
            self.state = 42
            self.levels()
            self.state = 43
            self.unlocks()
            self.state = 44
            self.achievements()
            self.state = 45
            self.match(SaveDataParser.ENDER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Date_timeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DATE(self):
            return self.getToken(SaveDataParser.DATE, 0)

        def TIME(self):
            return self.getToken(SaveDataParser.TIME, 0)

        def getRuleIndex(self):
            return SaveDataParser.RULE_date_time

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDate_time" ):
                listener.enterDate_time(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDate_time" ):
                listener.exitDate_time(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDate_time" ):
                return visitor.visitDate_time(self)
            else:
                return visitor.visitChildren(self)




    def date_time(self):

        localctx = SaveDataParser.Date_timeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_date_time)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 47
            self.match(SaveDataParser.DATE)
            self.state = 48
            self.match(SaveDataParser.TIME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DurationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self):
            return self.getTypedRuleContext(SaveDataParser.ValueContext,0)


        def DURATION_UNIT(self):
            return self.getToken(SaveDataParser.DURATION_UNIT, 0)

        def getRuleIndex(self):
            return SaveDataParser.RULE_duration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDuration" ):
                listener.enterDuration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDuration" ):
                listener.exitDuration(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDuration" ):
                return visitor.visitDuration(self)
            else:
                return visitor.visitChildren(self)




    def duration(self):

        localctx = SaveDataParser.DurationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_duration)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 50
            self.value()
            self.state = 51
            self.match(SaveDataParser.DURATION_UNIT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class InventoryContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INVENTORY_HEADER(self):
            return self.getToken(SaveDataParser.INVENTORY_HEADER, 0)

        def QUANTUM_FUSER(self):
            return self.getToken(SaveDataParser.QUANTUM_FUSER, 0)

        def value(self):
            return self.getTypedRuleContext(SaveDataParser.ValueContext,0)


        def getRuleIndex(self):
            return SaveDataParser.RULE_inventory

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInventory" ):
                listener.enterInventory(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInventory" ):
                listener.exitInventory(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitInventory" ):
                return visitor.visitInventory(self)
            else:
                return visitor.visitChildren(self)




    def inventory(self):

        localctx = SaveDataParser.InventoryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_inventory)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 53
            self.match(SaveDataParser.INVENTORY_HEADER)
            self.state = 54
            self.match(SaveDataParser.QUANTUM_FUSER)
            self.state = 55
            self.value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class GatesContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def GATES_HEADER(self):
            return self.getToken(SaveDataParser.GATES_HEADER, 0)

        def gate(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SaveDataParser.GateContext)
            else:
                return self.getTypedRuleContext(SaveDataParser.GateContext,i)


        def getRuleIndex(self):
            return SaveDataParser.RULE_gates

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterGates" ):
                listener.enterGates(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitGates" ):
                listener.exitGates(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitGates" ):
                return visitor.visitGates(self)
            else:
                return visitor.visitChildren(self)




    def gates(self):

        localctx = SaveDataParser.GatesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_gates)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 57
            self.match(SaveDataParser.GATES_HEADER)
            self.state = 61
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==18 or _la==22:
                self.state = 58
                self.gate()
                self.state = 63
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class GateContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME_STD(self):
            return self.getToken(SaveDataParser.NAME_STD, 0)

        def COMBINED_GATE(self):
            return self.getToken(SaveDataParser.COMBINED_GATE, 0)

        def combined_gate(self):
            return self.getTypedRuleContext(SaveDataParser.Combined_gateContext,0)


        def getRuleIndex(self):
            return SaveDataParser.RULE_gate

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterGate" ):
                listener.enterGate(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitGate" ):
                listener.exitGate(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitGate" ):
                return visitor.visitGate(self)
            else:
                return visitor.visitChildren(self)




    def gate(self):

        localctx = SaveDataParser.GateContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_gate)
        try:
            self.state = 67
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [22]:
                self.enterOuterAlt(localctx, 1)
                self.state = 64
                self.match(SaveDataParser.NAME_STD)
                pass
            elif token in [18]:
                self.enterOuterAlt(localctx, 2)
                self.state = 65
                self.match(SaveDataParser.COMBINED_GATE)
                self.state = 66
                self.combined_gate()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Combined_gateContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self):
            return self.getToken(SaveDataParser.TEXT, 0)

        def value(self):
            return self.getTypedRuleContext(SaveDataParser.ValueContext,0)


        def background_gate(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SaveDataParser.Background_gateContext)
            else:
                return self.getTypedRuleContext(SaveDataParser.Background_gateContext,i)


        def LIST_SEPARATOR(self, i:int=None):
            if i is None:
                return self.getTokens(SaveDataParser.LIST_SEPARATOR)
            else:
                return self.getToken(SaveDataParser.LIST_SEPARATOR, i)

        def getRuleIndex(self):
            return SaveDataParser.RULE_combined_gate

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCombined_gate" ):
                listener.enterCombined_gate(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCombined_gate" ):
                listener.exitCombined_gate(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitCombined_gate" ):
                return visitor.visitCombined_gate(self)
            else:
                return visitor.visitChildren(self)




    def combined_gate(self):

        localctx = SaveDataParser.Combined_gateContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_combined_gate)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 69
            self.match(SaveDataParser.T__0)
            self.state = 70
            self.match(SaveDataParser.TEXT)
            self.state = 71
            self.value()
            self.state = 72
            self.match(SaveDataParser.T__1)
            self.state = 73
            self.background_gate()
            self.state = 78
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==32:
                self.state = 74
                self.match(SaveDataParser.LIST_SEPARATOR)
                self.state = 75
                self.background_gate()
                self.state = 80
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 81
            self.match(SaveDataParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Background_gateContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME_STD(self):
            return self.getToken(SaveDataParser.NAME_STD, 0)

        def value(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SaveDataParser.ValueContext)
            else:
                return self.getTypedRuleContext(SaveDataParser.ValueContext,i)


        def LIST_SEPARATOR(self, i:int=None):
            if i is None:
                return self.getTokens(SaveDataParser.LIST_SEPARATOR)
            else:
                return self.getToken(SaveDataParser.LIST_SEPARATOR, i)

        def getRuleIndex(self):
            return SaveDataParser.RULE_background_gate

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBackground_gate" ):
                listener.enterBackground_gate(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBackground_gate" ):
                listener.exitBackground_gate(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitBackground_gate" ):
                return visitor.visitBackground_gate(self)
            else:
                return visitor.visitChildren(self)




    def background_gate(self):

        localctx = SaveDataParser.Background_gateContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_background_gate)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 83
            self.match(SaveDataParser.NAME_STD)
            self.state = 84
            self.match(SaveDataParser.T__3)
            self.state = 85
            self.value()
            self.state = 90
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==32:
                self.state = 86
                self.match(SaveDataParser.LIST_SEPARATOR)
                self.state = 87
                self.value()
                self.state = 92
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 93
            self.match(SaveDataParser.T__4)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LevelsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LEVELS_HEADER(self):
            return self.getToken(SaveDataParser.LEVELS_HEADER, 0)

        def level(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SaveDataParser.LevelContext)
            else:
                return self.getTypedRuleContext(SaveDataParser.LevelContext,i)


        def getRuleIndex(self):
            return SaveDataParser.RULE_levels

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLevels" ):
                listener.enterLevels(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLevels" ):
                listener.exitLevels(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLevels" ):
                return visitor.visitLevels(self)
            else:
                return visitor.visitChildren(self)




    def levels(self):

        localctx = SaveDataParser.LevelsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_levels)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 95
            self.match(SaveDataParser.LEVELS_HEADER)
            self.state = 99
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==24:
                self.state = 96
                self.level()
                self.state = 101
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LevelContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME_SPECIAL(self):
            return self.getToken(SaveDataParser.NAME_SPECIAL, 0)

        def date_time(self):
            return self.getTypedRuleContext(SaveDataParser.Date_timeContext,0)


        def duration(self):
            return self.getTypedRuleContext(SaveDataParser.DurationContext,0)


        def score(self):
            return self.getTypedRuleContext(SaveDataParser.ScoreContext,0)


        def getRuleIndex(self):
            return SaveDataParser.RULE_level

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLevel" ):
                listener.enterLevel(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLevel" ):
                listener.exitLevel(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLevel" ):
                return visitor.visitLevel(self)
            else:
                return visitor.visitChildren(self)




    def level(self):

        localctx = SaveDataParser.LevelContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_level)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 102
            self.match(SaveDataParser.NAME_SPECIAL)
            self.state = 103
            self.match(SaveDataParser.T__5)
            self.state = 104
            self.date_time()
            self.state = 105
            self.duration()
            self.state = 106
            self.score()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ScoreContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self):
            return self.getTypedRuleContext(SaveDataParser.ValueContext,0)


        def getRuleIndex(self):
            return SaveDataParser.RULE_score

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterScore" ):
                listener.enterScore(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitScore" ):
                listener.exitScore(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitScore" ):
                return visitor.visitScore(self)
            else:
                return visitor.visitChildren(self)




    def score(self):

        localctx = SaveDataParser.ScoreContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_score)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 108
            self.match(SaveDataParser.T__6)
            self.state = 109
            self.match(SaveDataParser.T__7)
            self.state = 110
            self.value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def VALUE(self):
            return self.getToken(SaveDataParser.VALUE, 0)

        def getRuleIndex(self):
            return SaveDataParser.RULE_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterValue" ):
                listener.enterValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitValue" ):
                listener.exitValue(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitValue" ):
                return visitor.visitValue(self)
            else:
                return visitor.visitChildren(self)




    def value(self):

        localctx = SaveDataParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_value)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 112
            self.match(SaveDataParser.VALUE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class UnlocksContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def UNLOCKS_HEADER(self):
            return self.getToken(SaveDataParser.UNLOCKS_HEADER, 0)

        def unlock(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SaveDataParser.UnlockContext)
            else:
                return self.getTypedRuleContext(SaveDataParser.UnlockContext,i)


        def getRuleIndex(self):
            return SaveDataParser.RULE_unlocks

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUnlocks" ):
                listener.enterUnlocks(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUnlocks" ):
                listener.exitUnlocks(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitUnlocks" ):
                return visitor.visitUnlocks(self)
            else:
                return visitor.visitChildren(self)




    def unlocks(self):

        localctx = SaveDataParser.UnlocksContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_unlocks)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 114
            self.match(SaveDataParser.UNLOCKS_HEADER)
            self.state = 118
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==22:
                self.state = 115
                self.unlock()
                self.state = 120
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class UnlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME_STD(self):
            return self.getToken(SaveDataParser.NAME_STD, 0)

        def date_time(self):
            return self.getTypedRuleContext(SaveDataParser.Date_timeContext,0)


        def getRuleIndex(self):
            return SaveDataParser.RULE_unlock

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUnlock" ):
                listener.enterUnlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUnlock" ):
                listener.exitUnlock(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitUnlock" ):
                return visitor.visitUnlock(self)
            else:
                return visitor.visitChildren(self)




    def unlock(self):

        localctx = SaveDataParser.UnlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_unlock)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 121
            self.match(SaveDataParser.NAME_STD)
            self.state = 122
            self.match(SaveDataParser.T__5)
            self.state = 123
            self.date_time()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AchievementsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ACHIEVEMENTS_HEADER(self):
            return self.getToken(SaveDataParser.ACHIEVEMENTS_HEADER, 0)

        def achievement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SaveDataParser.AchievementContext)
            else:
                return self.getTypedRuleContext(SaveDataParser.AchievementContext,i)


        def getRuleIndex(self):
            return SaveDataParser.RULE_achievements

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAchievements" ):
                listener.enterAchievements(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAchievements" ):
                listener.exitAchievements(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAchievements" ):
                return visitor.visitAchievements(self)
            else:
                return visitor.visitChildren(self)




    def achievements(self):

        localctx = SaveDataParser.AchievementsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_achievements)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 125
            self.match(SaveDataParser.ACHIEVEMENTS_HEADER)
            self.state = 129
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==22:
                self.state = 126
                self.achievement()
                self.state = 131
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AchievementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME_STD(self):
            return self.getToken(SaveDataParser.NAME_STD, 0)

        def date_time(self):
            return self.getTypedRuleContext(SaveDataParser.Date_timeContext,0)


        def score(self):
            return self.getTypedRuleContext(SaveDataParser.ScoreContext,0)


        def value(self):
            return self.getTypedRuleContext(SaveDataParser.ValueContext,0)


        def getRuleIndex(self):
            return SaveDataParser.RULE_achievement

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAchievement" ):
                listener.enterAchievement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAchievement" ):
                listener.exitAchievement(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAchievement" ):
                return visitor.visitAchievement(self)
            else:
                return visitor.visitChildren(self)




    def achievement(self):

        localctx = SaveDataParser.AchievementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_achievement)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 132
            self.match(SaveDataParser.NAME_STD)
            self.state = 133
            self.match(SaveDataParser.T__5)
            self.state = 134
            self.date_time()
            self.state = 135
            self.score()
            self.state = 136
            self.match(SaveDataParser.T__8)
            self.state = 137
            self.match(SaveDataParser.T__9)
            self.state = 138
            self.value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class IntegerContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DIGIT(self):
            return self.getToken(SaveDataParser.DIGIT, 0)

        def HALLWAY_ID(self):
            return self.getToken(SaveDataParser.HALLWAY_ID, 0)

        def INTEGER(self):
            return self.getToken(SaveDataParser.INTEGER, 0)

        def getRuleIndex(self):
            return SaveDataParser.RULE_integer

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInteger" ):
                listener.enterInteger(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInteger" ):
                listener.exitInteger(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitInteger" ):
                return visitor.visitInteger(self)
            else:
                return visitor.visitChildren(self)




    def integer(self):

        localctx = SaveDataParser.IntegerContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_integer)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 140
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 70574902607872) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Complex_numberContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IMAG_NUMBER(self):
            return self.getToken(SaveDataParser.IMAG_NUMBER, 0)

        def SIGN(self, i:int=None):
            if i is None:
                return self.getTokens(SaveDataParser.SIGN)
            else:
                return self.getToken(SaveDataParser.SIGN, i)

        def integer(self):
            return self.getTypedRuleContext(SaveDataParser.IntegerContext,0)


        def FLOAT(self):
            return self.getToken(SaveDataParser.FLOAT, 0)

        def getRuleIndex(self):
            return SaveDataParser.RULE_complex_number

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComplex_number" ):
                listener.enterComplex_number(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComplex_number" ):
                listener.exitComplex_number(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitComplex_number" ):
                return visitor.visitComplex_number(self)
            else:
                return visitor.visitChildren(self)




    def complex_number(self):

        localctx = SaveDataParser.Complex_numberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_complex_number)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 143
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==40:
                self.state = 142
                self.match(SaveDataParser.SIGN)


            self.state = 154
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [39]:
                self.state = 145
                self.match(SaveDataParser.IMAG_NUMBER)
                pass
            elif token in [36, 37, 38, 46]:
                self.state = 148
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [36, 37, 46]:
                    self.state = 146
                    self.integer()
                    pass
                elif token in [38]:
                    self.state = 147
                    self.match(SaveDataParser.FLOAT)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 152
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==40:
                    self.state = 150
                    self.match(SaveDataParser.SIGN)
                    self.state = 151
                    self.match(SaveDataParser.IMAG_NUMBER)


                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








# stores constants used in grammar files


class MapGrammarConfig:
    @staticmethod
    def name_prefix() -> str:
        return "Name = "

    @staticmethod
    def description_prefix() -> str:
        return "Description = "


class PuzzleGrammarConfig:
    @staticmethod
    def boss_code() -> str:
        return "code"


class SaveGrammarConfig:
    @staticmethod
    def combined_prefix() -> str:
        return "combined"



import base64
import datetime
import enum
import math
from typing import Optional, Any, Tuple, List

import numpy as np

__DEFAULT_CHARACTER = " "


def cur_datetime() -> datetime.datetime:
    return datetime.datetime.now()


def time_diff(time1: datetime, time2: datetime) -> Tuple[int, str]:
    # '2023-09-20 20:55:54.036295'
    diff = abs(time1 - time2)
    return diff.seconds, diff


def datetime2str(date_time: datetime) -> str:
    return date_time.strftime('%dd%mm%Yy %H:%M:%S')


def is_power_of_2(n: int):
    """
    Checks if a given number is a power of 2.
    :param n: the number we want to check
    :return: True if n is a power of 2, False otherwise
    """
    if n == 0:
        return False
    return (n & (n - 1)) == 0


def clamp(val: int, minv: int, maxv: int) -> int:
    """
    Returns val if minv <= val <= maxv is True. Else either returns minv or maxv, depending on which one is closer to
    val.

    :param val: the value to clamp
    :param minv: the minimum value val should be clamped to
    :param maxv: the maximum value val should be clamped to
    :return: minv if val < minv, maxv if val > maxv, else val (minv <= val <= maxv)
    """
    return max(minv, min(val, maxv))


def rad2deg(rad: float) -> float:
    return rad * 180 / math.pi


def complex2string(val: complex, decimals: int) -> str:
    val = np.round(val, decimals)
    if val.imag == 0:
        text = f"{val.real:g}"  # g turns 0.0 to 0
    elif val.real == 0:
        text = f"{val.imag:g}j"
    else:
        if val.real != 0 and val.real != 1 and val.real != -1:
            real = f"{val.real:g}"
            if real[0] == "0":
                real = real[1:]  # remove the redundant "0" in front of "0.x", resulting in ".x"
            else:
                real = real[0] + real[2:]  # remove the redundant "0" after the sign, resulting in "-.x"
        else:
            real = str(val.real)  # real is now either "0" or "1" or "-1"
        if val.imag == 1:
            text = f"{real}+j"
        elif val.imag == -1:
            text = f"{real}-j"
        else:
            imag = f"{val.imag:g}"
            if imag[0] == "0":  # remove the redundant "0" in front of "0.x", resulting in ".x"
                imag = f"+{imag[1:]}"
            else:  # remove the redundant "0" after the sign, resulting in "-.x"
                imag = imag[0] + imag[2:]
            text = f"{real}{imag}j"
    # skip "-" in front if the text starts with "-0" and the value is actually 0 (so no more comma)
    if text.startswith("-0") and (len(text) == 2 or len(text) > 2 and text[2] != "."):
        text = text[1:]  # I think this can never happen since we throw away 0 real- or imag-parts  # IT CAN HAPPEN!
    return text


def to_binary_string(num: int, digits: Optional[int] = None, msb: bool = True) -> str:
    """
    Converts a given integer into its binary representation as string.

    :param num: the number we want to convert to a binary string
    :param digits: number of digits the string should show (= length), if None it is calculated to fit num
    :param msb: whether to place the most or least significant bit(s) first
    :return:
    """
    if digits is None:
        digits = int(math.ceil(math.log2(num)))
    str_rep = ""
    i = 0
    while True:
        str_rep += str(num % 2)
        num = int(num / 2)
        i += 1
        if digits <= i:
            break
        elif digits < 0 and num <= 0:
            break
    if msb:
        return str_rep[::-1]  # reverse the string
    else:
        return str_rep


def compute_centering(text: str, line_width: int, uneven_left: bool = True, character: str = __DEFAULT_CHARACTER) \
        -> Tuple[str, str]:
    """
    Computes a prefix and suffix for the given text in such a way that it is as centered as possible in a line with the
    given width. This prefix and suffix will consist only of the given character (repeated for centering). If the number
    of characters to add is uneven (i.e. number of prepended and appended characters cannot be the same) the flag
    uneven_left decides whether to prepend (= left, default) or append (= right) should be bigger.

    :param text: the text to center in the line
    :param line_width: width of the line the text should be centered in
    :param uneven_left: whether to prepend or append one character more in case of imbalance, defaults to True
    :param character: the character to add, defaults to whitespace " "
    :return: prefix, suffix, such that text is centered
    """
    if line_width <= len(text):
        return "", ""
    diff = line_width - len(text)
    half1 = int(diff / 2)
    half2 = diff - half1

    if uneven_left:
        return half2 * character, half1 * character
    else:
        return half1 * character, half2 * character


def center_string(text: str, line_width: int, uneven_left: bool = True, character: str = __DEFAULT_CHARACTER) -> str:
    """
    Prepends and appends the given character to the text in such a way that the text is centered as good as possible
    in a line with the given line width. If the number of characters to add is uneven (i.e. number of prepended and
    appended characters cannot be the same) the flag uneven_left decides whether to prepend (= left, default) or
    append (= right) should be bigger.

    :param text: the text to center in the line
    :param line_width: width of the line the text should be centered in
    :param uneven_left: whether to prepend or append one character more in case of imbalance, defaults to True
    :param character: the character to add, defaults to whitespace " "
    :return: centered version of text
    """
    prefix, suffix = compute_centering(text, line_width, uneven_left, character)
    return prefix + text + suffix


def align_string(text: str, line_width: int, left: bool = True, character: str = __DEFAULT_CHARACTER) -> str:
    """
    Tries to align text in a line by adding character to the left or right until text has the length of the given line
    width. If text is not shorter than line width it is immediately returned without any changes.

    :param text: the text to align
    :param line_width: the length text should have in the end
    :param left: whether the text should be left-aligned (True) or right-aligned (false)
    :param character: the character to add as alignment, defaults to whitespace
    :return: a left or right aligned version of text
    """
    if line_width <= len(text):
        return text
    diff = line_width - len(text)
    if left:
        return text + diff * character
    else:
        return diff * character + text


def align_lines(lines: List[str], titles: List[str], separators: List[str],
                left: bool = True, character: str = __DEFAULT_CHARACTER) -> List[str]:
    assert len(lines) == len(titles) == len(separators), \
        f"Lengths are different: {len(lines)} =?= {len(titles)} =?= {len(separators)}"

    # 1) align the lines
    max_line_width = max([len(line) for line in lines])
    lines = [align_string(line, max_line_width, left, character) for line in lines]

    # 2) align the titles with corresponding space between their lines
    # make sure the titles are extended such that the first/last character is placed in the same column
    max_line_width = max([len(title) for title in titles])
    titles = [align_string(title, max_line_width, left, character) for title in titles]

    # 3) combine titles with separators and lines
    if left:
        return [titles[i] + separators[i] + lines[i] for i in range(len(lines))]
    else:
        return [lines[i] + separators[i] + titles[i] for i in range(len(lines))]


def enum_string(value: enum.Enum, skip_type_prefix: bool = True) -> str:
    text = str(value)
    if skip_type_prefix and "." in text:
        index = text.index(".")
        return text[index + 1:]
    return text


def enum_from_string(enum_type, text: str) -> Optional:
    # remove potential AreaType-prefix
    type_string = str(enum_type)
    # the enum_type is written like this: "<enum '{ActualEnumName}'>"
    if "enum" not in type_string: return None
    type_start = type_string.index("'") + 1
    type_end = type_string.index("'", type_start)
    type_string = type_string[type_start:type_end]
    if text.startswith(type_string):
        text = text[len(type_string):]
    # iterate through all enum values and find the one with the same name
    for val in enum_type:
        if val.name == text:
            return val
    return None


def my_str(value: Any) -> str:
    """
    Just like str() but with special treatment for Enums.

    :param value:
    :return: string representation of the given value
    """
    if isinstance(value, enum.Enum):
        return enum_string(value)
    else:
        return str(value)


def int_to_fixed_len_str(value: int, length: int) -> str:
    """
    Returns a formatted string like "007" if value=7 and length=3.
    If value is too big (e.g., it has more than $length digits) length is ignored and value is just transformed to its
    normal string representation (e.g., value=123 and length=1 returns "123").

    :param value: the integer we want to represent
    :param length: the length of the formatted string
    :return: 0-padded string representation of $value with a length of $length or len(value)
    """
    if len(str(value)) > length:
        return str(value)
    return f"{{0:0{length}d}}".format(value)


def num_to_letter(num: int, start_uppercase: bool = True) -> str:
    if num < 0: return "?"

    if num < 26:
        # below 26 means we can start normally (either uppercase at @ or lowercase at `
        start_char = "@" if start_uppercase else "`"
    else:
        # above 26 (and below 52, else it will be normalized below) means we have to start with the other case
        num -= 26
        start_char = "`" if start_uppercase else "@"

    num = num % 26  # normalize to one of the 26 letters
    return chr(ord(start_char) + num)


def simple_encode(key: str, clear: str) -> str:
    enc = []
    for i in range(len(clear)):
        key_c = key[i % len(key)]
        enc_c = chr((ord(clear[i]) + ord(key_c)) % 256)
        enc.append(enc_c)
    enc = base64.urlsafe_b64encode(bytearray("".join(enc), "utf-8"))
    return enc.decode("utf-8")  # convert bytes back to str


def simple_decode(key: str, enc: str) -> str:
    dec = []
    enc = base64.urlsafe_b64decode(enc)
    enc = enc.decode("utf-8")   # convert bytes back to str
    for i in range(len(enc)):
        key_c = key[i % len(key)]
        dec_c = chr((256 + ord(enc[i]) - ord(key_c)) % 256)
        dec.append(dec_c)
    return "".join(dec)


def open_folder(path: str):
    import os
    import platform
    import subprocess

    try:
        # worked on Windows, Ubuntu and macOS
        if platform.system() == "Windows":
            os.startfile(path)
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", path])
        else:
            subprocess.Popen(["xdg-open", path])
    except Exception as ex:
        raise ex
